#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Utility classes and functions."""

from typing import Tuple

import numpy as np
import pinocchio as pin

from .exceptions import ConfigurationError, PinkError


def custom_configuration_vector(robot: pin.Model, **kwargs) -> np.ndarray:
    """Generate a configuration vector where named joints have specific values.

    Args:
        robot: Robot model.
        kwargs: Custom values for joint coordinates.

    Returns:
        Configuration vector where named joints have the values specified in
        keyword arguments, and other joints have their neutral value.
    """
    q = pin.neutral(robot.model)
    for name, value in kwargs.items():
        joint_id = robot.model.getJointId(name)
        joint = robot.model.joints[joint_id]
        value = np.array(value).flatten()
        if value.shape[0] != joint.nq:
            raise ConfigurationError(
                f"Joint '{name}' has {joint.nq=} but is set to {value.shape=}"
            )
        q[joint.idx_q : joint.idx_q + joint.nq] = value
    return q


def get_root_joint_dim(model: pin.Model) -> Tuple[int, int]:
    """Count configuration and tangent dimensions of the root joint, if any.

    Args:
        model: Robot model.

    Returns:
        nq: Number of configuration dimensions.
        nv: Number of tangent dimensions.
    """
    if model.existJointName("root_joint"):
        root_joint_id = model.getJointId("root_joint")
        root_joint = model.joints[root_joint_id]
        return root_joint.nq, root_joint.nv
    return 0, 0


def get_joint_idx(model: pin.Model, joint_name: str) -> Tuple[int, int]:
    """Get joint index in the configuration and tangent space.

    Args:
        model: Robot model.
        joint_name: Joint name.

    Returns:
        idx_q: Joint idx in configuration space.
        idx_v: Joint idx in tangent space.
    """
    if model.existJointName(joint_name):
        joint_id = model.getJointId(joint_name)
        joint = model.joints[joint_id]
        return joint.idx_q, joint.idx_v
    raise PinkError(
        f"cannot find the joint index corresponding to joint {joint_name}"
    )


class VectorSpace:
    """Wrapper to refer to a vector space and its characteristic matrices."""

    __eye: np.ndarray
    __ones: np.ndarray
    __zeros: np.ndarray

    def __init__(self, dim: int):
        """Create new vector space description.

        Args:
            dim: Dimension.
        """
        eye = np.eye(dim)
        ones = np.ones(dim)
        zeros = np.zeros(dim)
        eye.setflags(write=False)
        ones.setflags(write=False)
        zeros.setflags(write=False)
        self.__eye = eye
        self.__ones = ones
        self.__zeros = zeros

    @property
    def eye(self) -> np.ndarray:
        """Identity matrix from and to the vector space."""
        return self.__eye

    @property
    def ones(self) -> np.ndarray:
        """Vector full of ones, dimension of the space."""
        return self.__ones

    @property
    def zeros(self) -> np.ndarray:
        """Zero vector of the space."""
        return self.__zeros


def process_collision_pairs(
    model: pin.Model, collision_model: pin.GeometryModel, srdf_path: str = ""
) -> pin.GeometryData:
    """Process collision pairs.

    Args:
        model: robot model.
        collision_model: Collision model of the robot.
        srdf_path: Path to the SRDF file, which used to exclude collision
            pairs. Defaults to empty string, meaning no collision pairs are
            excluded.

    Returns:
        collision_data: Collision data, generated after updating
            collision_model.
    """
    collision_model.addAllCollisionPairs()
    if srdf_path != "":
        pin.removeCollisionPairs(model, collision_model, srdf_path)

    # Collision models have been modified => re-generate corresponding data.
    collision_data = pin.GeometryData(collision_model)

    # Enable contact detection for avoiding Nans at collisions
    collision_data.enable_contact = True

    return collision_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""Inverse kinematics for articulated robot models, based on Pinocchio."""

from .configuration import Configuration
from .exceptions import PinkError
from .solve_ik import build_ik, solve_ik
from .tasks import (
    FrameTask,
    JointCouplingTask,
    JointVelocityTask,
    LinearHolonomicTask,
    PostureTask,
    Task,
)
from .utils import custom_configuration_vector

__version__ = "3.4.0"

__all__ = [
    "Configuration",
    "FrameTask",
    "JointCouplingTask",
    "JointVelocityTask",
    "LinearHolonomicTask",
    "PinkError",
    "PostureTask",
    "Task",
    "build_ik",
    "custom_configuration_vector",
    "solve_ik",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers

import pink
from pink import solve_ik
from pink.tasks import FrameTask, RollingTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    )


if __name__ == "__main__":
    robot = load_robot_description(
        "upkie_description",
        root_joint=pin.JointModelFreeFlyer(),
    )
    visualizer = start_meshcat_visualizer(robot)

    base_task = FrameTask(
        "base",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
    )
    left_wheel_rolling = RollingTask(
        "left_wheel",
        floor_frame="universe",
        wheel_radius=0.06,
        cost=10.0,
    )
    right_wheel_rolling = RollingTask(
        "right_wheel",
        floor_frame="universe",
        wheel_radius=0.06,
        cost=10.0,
    )
    left_wheel_position = FrameTask(
        "left_wheel",
        position_cost=1.0,
        orientation_cost=0.0,
    )
    right_wheel_position = FrameTask(
        "right_wheel",
        position_cost=1.0,
        orientation_cost=0.0,
    )

    init_x = -2.0  # in [m]
    q_init = robot.q0.copy()
    q_init[0] = init_x  # world x-axis
    q_init[2] = 0.56  # world z-axis
    configuration = pink.Configuration(robot.model, robot.data, q_init)
    base_task.set_target_from_configuration(configuration)
    left_wheel_position.set_target_from_configuration(configuration)
    right_wheel_position.set_target_from_configuration(configuration)
    visualizer.display(configuration.q)

    base_target = base_task.transform_target_to_world
    left_wheel_target = left_wheel_position.transform_target_to_world
    right_wheel_target = right_wheel_position.transform_target_to_world

    viewer = visualizer.viewer
    meshcat_shapes.frame(viewer["base"], opacity=1.0)
    meshcat_shapes.frame(viewer["base_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["left_wheel_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["right_wheel_target"], opacity=0.5)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "proxqp" in qpsolvers.available_solvers:
        solver = "proxqp"

    rate = RateLimiter(frequency=200.0, warn=False)
    dt = rate.period
    t = 0.0  # [s]

    while True:
        # Update base task target
        base_x = 0.1 * t + init_x
        base_target.translation[0] = base_x
        left_wheel_target.translation[0] = base_x + 0.1 * np.sin(t)
        right_wheel_target.translation[0] = base_x - 0.1 * np.sin(t)

        # Update visualization frames
        viewer["base_target"].set_transform(base_target.np)
        viewer["left_wheel_target"].set_transform(left_wheel_target.np)
        viewer["right_wheel_target"].set_transform(right_wheel_target.np)
        viewer["base"].set_transform(
            configuration.get_transform_frame_to_world("base").np
        )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(
            configuration,
            tasks=[base_task, left_wheel_position, right_wheel_position],
            dt=dt,
            solver=solver,
            damping=1e-3,
            constraints=[left_wheel_rolling, right_wheel_rolling],
        )
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        visualizer.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0

"""Move the end-effector of a UR10 arm to a prescribed target."""

import numpy as np
import pinocchio as pin
import pinocchio
import qpsolvers

import pink
from pink.tasks import FrameTask

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    ) from exc

# IK parameters
dt = 1e-2
stop_thres = 1e-8

if __name__ == "__main__":
    robot = load_robot_description("ur10_description")

    # Frame details
    joint_name = robot.model.names[-1]
    parent_joint = robot.model.getJointId(joint_name)
    parent_frame = robot.model.getFrameId(joint_name)
    placement = pin.SE3.Identity()

    FRAME_NAME = "ee_frame"
    ee_frame = robot.model.addFrame(
        pin.Frame(
            FRAME_NAME,
            parent_joint,
            parent_frame,
            placement,
            pin.FrameType.OP_FRAME,
        )
    )
    robot.data = pin.Data(robot.model)
    low = robot.model.lowerPositionLimit
    high = robot.model.upperPositionLimit
    robot.q0 = pin.neutral(robot.model)

    # Task details
    np.random.seed(0)
    q_final = np.array(
        [
            np.random.uniform(low=low[i], high=high[i], size=(1,))[0]
            for i in range(6)
        ]
    )
    pin.forwardKinematics(robot.model, robot.data, q_final)
    target_pose = robot.data.oMi[parent_joint]
    ee_task = FrameTask(FRAME_NAME, [1.0, 1.0, 1.0], [1.0, 1.0, 1.0])
    ee_task.set_target(target_pose)

    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    error_norm = np.linalg.norm(ee_task.compute_error(configuration))
    print(f"Starting from {error_norm = :.2}")
    print(f"Desired precision is error_norm < {stop_thres}")

    nb_steps = 0
    while error_norm > stop_thres:
        dv = pink.solve_ik(
            configuration,
            tasks=[ee_task],
            dt=dt,
            damping=1e-8,
            solver=(
                "daqp"
                if "daqp" in qpsolvers.available_solvers
                else qpsolvers.available_solvers[0]
            ),
        )
        q_out = pin.integrate(robot.model, configuration.q, dv * dt)
        configuration = pink.Configuration(robot.model, robot.data, q_out)
        pin.updateFramePlacements(robot.model, robot.data)
        error_norm = np.linalg.norm(ee_task.compute_error(configuration))
        nb_steps += 1

    print(f"Terminated after {nb_steps} steps with {error_norm = :.2}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Two iiwa14-s with full-body self-collision avoidance using hpp-fcl."""

import os

import numpy as np
import pinocchio as pin
import qpsolvers

import meshcat_shapes
import pink
from pink import solve_ik
from pink.barriers import SelfCollisionBarrier
from pink.tasks import FrameTask, PostureTask
from pink.utils import process_collision_pairs
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.iiwa14_description import (
        PACKAGE_PATH,
        REPOSITORY_PATH,
    )
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    ) from exc


def prefix_frames(
    model: pin.Model,
    visual_model: pin.GeometryModel,
    geometry_model: pin.GeometryModel,
    prefix: str,
) -> None:
    for frame in model.frames:
        frame.name = f"{prefix}_{frame.name}"
    for i, name in enumerate(model.names):
        model.names[i] = f"{prefix}_{name}"
    for geom in visual_model.geometryObjects:
        geom.name = f"{prefix}_{geom.name}"
    for geom in collision_model.geometryObjects:
        geom.name = f"{prefix}_{geom.name}"


if __name__ == "__main__":
    # Empty model
    model, visual_model, collision_model = (
        pin.Model(),
        pin.GeometryModel(),
        pin.GeometryModel(),
    )
    urdf_path = os.path.join(
        PACKAGE_PATH, "urdf", "iiwa14_spheres_collision.urdf"
    )

    # === Left arm ====
    left_arm = pin.RobotWrapper.BuildFromURDF(
        urdf_path, package_dirs=[os.path.dirname(REPOSITORY_PATH)]
    )

    # Add prefix to frames, links and geons of the arm
    prefix_frames(
        left_arm.model, left_arm.visual_model, left_arm.collision_model, "left"
    )

    # Place left arm on the left to the origin
    left_arm_placement = pin.SE3.Identity()
    left_arm_placement.translation = np.array([0.0, 0.2, 0.0])

    # Add left arm's model, visual model and collision model to the main model
    _, visual_model = pin.appendModel(
        model,
        left_arm.model,
        visual_model,
        left_arm.visual_model,
        0,
        left_arm_placement,
    )
    model, collision_model = pin.appendModel(
        model,
        left_arm.model,
        collision_model,
        left_arm.collision_model,
        0,
        left_arm_placement,
    )

    # === Right arm ====
    right_arm = pin.RobotWrapper.BuildFromURDF(
        urdf_path, package_dirs=[os.path.dirname(REPOSITORY_PATH)]
    )

    # Add prefix to frames, links and geons of the arm
    prefix_frames(
        right_arm.model,
        right_arm.visual_model,
        right_arm.collision_model,
        "right",
    )

    # Place left arm on the left to the origin
    right_arm_placement = pin.SE3.Identity()
    right_arm_placement.translation = np.array([0.0, -0.2, 0.0])

    # Add left arm's model, visual model and collision model to the main model
    _, visual_model = pin.appendModel(
        model,
        right_arm.model,
        visual_model,
        right_arm.visual_model,
        0,
        right_arm_placement,
    )
    model, collision_model = pin.appendModel(
        model,
        right_arm.model,
        collision_model,
        right_arm.collision_model,
        0,
        right_arm_placement,
    )

    # Assemble to the robot
    robot = pin.RobotWrapper(
        model,
        collision_model=collision_model,
        visual_model=visual_model,
    )

    srdf_path = (
        os.path.dirname(os.path.realpath(__file__))
        + "/iiwa14_spheres_collision.srdf"
    )
    print(srdf_path)
    viz = start_meshcat_visualizer(robot)
    q_ref = np.zeros(robot.model.nq)

    # Collisions: processing collisions from urdf (include all) and srdf
    # (exclude specified) and updating collision model and creating
    # corresponding collision data
    robot.collision_data = process_collision_pairs(
        robot.model, robot.collision_model, srdf_path
    )

    configuration = pink.Configuration(
        robot.model,
        robot.data,
        q_ref,
        collision_model=robot.collision_model,  # for self-collision barrier
        collision_data=robot.collision_data,
    )

    # Pink tasks
    left_end_effector_task = FrameTask(
        "left_iiwa_link_7",
        position_cost=50.0,  # [cost] / [m]
        orientation_cost=10.0,  # [cost] / [rad]
    )
    right_end_effector_task = FrameTask(
        "right_iiwa_link_7",
        position_cost=50.0,  # [cost] / [m]
        orientation_cost=10.0,  # [cost] / [rad]
    )

    # Pink barriers
    collision_barrier = SelfCollisionBarrier(
        n_collision_pairs=len(robot.collision_model.collisionPairs),
        gain=20.0,
        safe_displacement_gain=1.0,
        d_min=0.05,
    )

    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )

    barriers = [collision_barrier]
    tasks = [left_end_effector_task, right_end_effector_task, posture_task]

    for task in tasks:
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    viewer = viz.viewer
    meshcat_shapes.frame(viewer["left_end_effector"], opacity=1.0)
    meshcat_shapes.frame(viewer["right_end_effector"], opacity=1.0)
    meshcat_shapes.frame(viewer["left_end_effector_target"], opacity=1.0)
    meshcat_shapes.frame(viewer["right_end_effector_target"], opacity=1.0)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "osqp" in qpsolvers.available_solvers:
        solver = "osqp"

    rate = RateLimiter(frequency=50.0)
    dt = rate.period
    t = 0.0  # [s]
    l_y_des = np.array([0.392, -0.392, 0.6])
    r_y_des = np.array([0.392, 0.392, 0.6])

    A = l_y_des.copy()
    B = r_y_des.copy()

    l_dy_des = np.zeros(3)
    r_dy_des = np.zeros(3)

    while True:
        # Make a sinusoidal trajectory between points A and B
        mu = (1 + np.cos(t)) / 2
        l_y_des[:] = (
            A + (B - A + 0.2 * np.array([0, 0, np.sin(mu * np.pi) ** 2])) * mu
        )
        r_y_des[:] = (
            B + (A - B + 0.2 * np.array([0, 0, -np.sin(mu * np.pi) ** 2])) * mu
        )

        left_end_effector_task.transform_target_to_world.translation = l_y_des
        right_end_effector_task.transform_target_to_world.translation = r_y_des

        # Update visualization frames
        viewer["left_end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                left_end_effector_task.frame
            ).np
        )
        viewer["right_end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                right_end_effector_task.frame
            ).np
        )
        viewer["left_end_effector_target"].set_transform(
            left_end_effector_task.transform_target_to_world.np
        )
        viewer["right_end_effector_target"].set_transform(
            right_end_effector_task.transform_target_to_world.np
        )

        velocity = solve_ik(
            configuration,
            tasks,
            dt,
            solver=solver,
            barriers=barriers,
            safety_break=False,
        )
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Domrachev Ivan, Simeon Nedelchev

"""Test fixture for the com task."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.exceptions import TargetNotSet
from pink.tasks import ComTask


class TestComTask(unittest.TestCase):
    """Test consistency of the com task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        self.configuration = Configuration(robot.model, robot.data, robot.q0)

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        task = ComTask(cost=1.0)
        self.assertTrue("cost=" in repr(task))
        self.assertTrue("gain=" in repr(task))
        self.assertTrue("target_com=" in repr(task))

    def test_target_not_set(self):
        """Raise an exception when the target is not set."""
        task = ComTask(cost=1.0)
        with self.assertRaises(TargetNotSet):
            task.compute_error(self.configuration)

    def test_set_target_from_configuration(self):
        """Check that target is set from its configuration value."""
        task = ComTask(cost=1.0)
        task.set_target_from_configuration(self.configuration)
        com = pin.centerOfMass(
            self.configuration.model,
            self.configuration.data,
            self.configuration.q,
        )
        self.assertTrue(np.allclose(com, task.target_com))

    def test_target_set_properly(self):
        """Return target properly once it's set."""
        task = ComTask(cost=1.0)
        com_task = np.random.rand(3)
        task.set_target(com_task)
        self.assertIsNotNone(task.target_com)
        if task.target_com is not None:  # help mypy
            self.assertTrue(np.allclose(com_task, task.target_com))

    def test_target_is_a_copy(self):
        """Target is saved as a copy, not a reference to the original."""
        task = ComTask(cost=1.0)
        com_des = np.random.rand(3)
        task.set_target(com_des)
        self.assertIsNotNone(task.target_com)

    def test_zero_error_when_target_at_body(self):
        """Error is zero when the target and body are at the same location."""
        task = ComTask(cost=1.0)
        task.set_target(
            pin.centerOfMass(self.configuration.model, self.configuration.data)
        )  # error == 0
        e = task.compute_error(self.configuration)
        self.assertLess(np.linalg.norm(e), 1e-10)

    def test_unit_cost_qp_objective(self):
        """A unit cost vector means the QP objective is (J^T J, -e^T J)."""
        task = ComTask(cost=1.0)
        task.set_target(
            pin.centerOfMass(self.configuration.model, self.configuration.data)
        )
        q_new = self.configuration.q.copy()
        q_new[1] += 1.0
        q_new[3] += 1.0
        q_new[5] += 1.0
        self.configuration.q = q_new
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        print
        self.assertTrue(np.allclose(e.T @ J, c))

    def test_zero_cost_same_as_disabling_task(self):
        """The task has no effect when its cost is zero."""
        task = ComTask(cost=0.0)
        task.set_target(
            pin.centerOfMass(self.configuration.model, self.configuration.data)
        )
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        qd = np.random.random(J.shape[1:])
        cost = qd.T @ H @ qd + c @ qd
        self.assertAlmostEqual(cost, 0.0)

    def test_cost_raises_exception(self):
        """The task raises an exception when its cost is
        negative and allows float, list and np.ndarray as input."""
        with self.assertRaises(AssertionError):
            task = ComTask(cost=-1.0)
        with self.assertRaises(AssertionError):
            task = ComTask(cost=[-1.0, -1.0, -1.0])
        with self.assertRaises(AssertionError):
            task = ComTask(cost=-np.ones(3))

        task = ComTask(cost=1.0)
        task.set_cost(cost=[1.0, 1.0, 1.0])
        task.set_cost(cost=np.ones(3))

    def test_jac_without_target(self):
        """The jacobian call should raise an exception
        when the target is not set"""

        task = ComTask(cost=1.0)
        with self.assertRaises(TargetNotSet):
            task.compute_jacobian(self.configuration)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Subset of acceleration-limited joints in a robot model."""

from typing import List, Optional, Tuple

import numpy as np
import pinocchio as pin

from ..exceptions import PinkError
from .limit import Limit


class AccelerationLimit(Limit):
    r"""Subset of acceleration-limited joints in a robot model.

    Acceleration limits consists of two parts: the expected :math:`|a| \leq
    a_{\mathrm{max}}`, but also the following term accounting for the "breaking
    distance" to configuration limits:

    .. math::

        -\sqrt{2 a_max (q \ominus q_{\min})} \leq a \leq \sqrt{2 a_max
        (q_{\max} \ominus q)}

    This additional inequality is detailed in [Flacco2015]_ as well as in
    [DelPrete2018]_.

    Attributes:
        Delta_q_prev: Latest displacement of the robot.
        a_max: Maximum acceleration vector for acceleration-limited joints.
        indices: Tangent indices corresponding to acceleration-limited joints.
        model: Robot model.
        projection_matrix: Projection from tangent space to subspace with
            acceleration-limited joints.
    """

    Delta_q_prev: np.ndarray
    a_max: np.ndarray
    indices: np.ndarray
    model: pin.Model
    projection_matrix: Optional[np.ndarray]

    def __init__(self, model: pin.Model, acceleration_limit: np.ndarray):
        """Initialize bounded tangent of a model.

        Args:
            model: robot model.
            acceleration_limit: Vector of acceleration limits, of dimension
                ``model.nv``.
        """
        acceleration_limit = acceleration_limit.flatten()
        if model.nv > 0 and acceleration_limit.shape[0] != model.nv:
            raise PinkError(f"{acceleration_limit.shape=} but {model.nv=}")

        has_acceleration_limit = np.logical_and(
            acceleration_limit < 1e20,
            acceleration_limit > 1e-10,
        )

        joints = [
            joint
            for joint in model.joints
            if joint.idx_v >= 0
            and has_acceleration_limit[
                slice(joint.idx_v, joint.idx_v + joint.nv)
            ].all()
        ]

        index_list: List[int] = []
        for joint in joints:
            index_list.extend(range(joint.idx_v, joint.idx_v + joint.nv))
        indices = np.array(index_list)
        indices.setflags(write=False)

        dim = len(indices)
        projection_matrix = np.eye(model.nv)[indices] if dim > 0 else None
        a_max = acceleration_limit[indices] if dim > 0 else np.empty(0)

        self.Delta_q_prev = np.zeros(dim)
        self.a_max = a_max
        self.indices = indices
        self.model = model
        self.projection_matrix = projection_matrix

    def set_last_integration(self, v_prev: np.ndarray, dt) -> None:
        """Set the latest velocity and the duration it was applied for.

        The goal of the low-acceleration task is to minimize the difference
        between the new velocity and the previous one.

        Args:
            v_prev: Latest integrated velocity.
            dt: Integration timestep in [s].
        """
        Delta_q_prev_full = v_prev * dt
        self.Delta_q_prev = Delta_q_prev_full[self.indices]

    def compute_qp_inequalities(
        self,
        q: np.ndarray,
        dt: float,
    ) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        r"""Compute inequalities for acceleration limits.

        Those limits are defined by:

        .. math::

            \Delta q_{\mathrm{prev}} - a_max \mathrm{d} t^2
            \leq \Delta q
            \leq \Delta q_{\mathrm{prev}} + a_max \mathrm{d} t^2

        where :math:`a_{max} \in {\cal T}` is the robot's acceleration limit
        vector (in the tangent space) and :math:`\Delta q \in T_q({\cal C})` is
        the displacement computed by the inverse kinematics, with :math:`\Delta
        q_{\mathrm{prev}}` the displacement from the previous iteration.

        Args:
            q: Robot configuration.
            dt: Integration timestep in [s].

        Returns:
            Pair :math:`(G, h)` representing the inequality constraint as
            :math:`G \Delta q \leq h`, or ``None`` if there is no limit.
        """
        if self.projection_matrix is None:  # no joint (thus checked for mypy)
            return None

        Delta_q_max = pin.difference(
            self.model, q, self.model.upperPositionLimit
        )[self.indices]
        Delta_q_min = pin.difference(
            self.model, self.model.lowerPositionLimit, q
        )[self.indices]
        dt_sq = dt * dt
        G = np.vstack([self.projection_matrix, -self.projection_matrix])
        h = np.hstack(
            [
                np.minimum(
                    self.a_max * dt_sq + self.Delta_q_prev,
                    np.sqrt(2 * self.a_max * Delta_q_max),
                ),
                np.minimum(
                    self.a_max * dt_sq - self.Delta_q_prev,
                    np.sqrt(2 * self.a_max * Delta_q_min),
                ),
            ]
        )
        return G, h



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

r"""Limits implemented as inequality constraints in the IK problem.

Kinematic limits derive from the :class:`Limit` base class. They compute QP
inequality constraints of the form:

.. math::

    G(q) \Delta q \leq h(q)

where :math:`q \in {\cal C}` is the robot's configuration and
:math:`\Delta q \in T_q({\cal C})` is the displacement in the tangent
space at :math:`q`.
"""

from .acceleration_limit import AccelerationLimit
from .configuration_limit import ConfigurationLimit
from .limit import Limit
from .velocity_limit import VelocityLimit

__all__ = [
    "AccelerationLimit",
    "ConfigurationLimit",
    "Limit",
    "VelocityLimit",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Tests for position barrier limit."""

import os
import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.iiwa14_description import PACKAGE_PATH, REPOSITORY_PATH

from pink import Configuration
from pink.barriers import SelfCollisionBarrier
from pink.exceptions import InvalidCollisionPairs, NegativeMinimumDistance
from pink.utils import process_collision_pairs


class TestSelfCollisionBarrier(unittest.TestCase):
    """Testing position barrier limit."""

    def setUp(self):
        """Set test fixture up."""
        urdf_path = os.path.join(
            PACKAGE_PATH, "urdf", "iiwa14_spheres_collision.urdf"
        )
        self.robot = pin.RobotWrapper.BuildFromURDF(
            urdf_path, package_dirs=[os.path.dirname(REPOSITORY_PATH)]
        )

        self.non_colliding_objects_pair_id = 5
        self.q0 = np.zeros(self.robot.model.nq)
        self.robot.collision_data = process_collision_pairs(
            self.robot.model, self.robot.collision_model
        )
        self.configuration = Configuration(
            self.robot.model,
            self.robot.data,
            self.q0,
            collision_model=self.robot.collision_model,
            collision_data=self.robot.collision_data,
        )

    def test_invalid_distance(self):
        """Raise an error when the robot body is not found."""
        with self.assertRaises(NegativeMinimumDistance):
            SelfCollisionBarrier(n_collision_pairs=10, d_min=-1)

    def test_invalid_collision_pairs(self):
        """Raise an error when the robot body is not found."""
        with self.assertRaises(InvalidCollisionPairs):
            SelfCollisionBarrier(n_collision_pairs=-1, d_min=0.02)
        with self.assertRaises(InvalidCollisionPairs):
            SelfCollisionBarrier(
                n_collision_pairs=len(
                    self.robot.collision_model.collisionPairs
                )
                + 5,
                d_min=0.02,
            ).compute_barrier(self.configuration)

    def test_dimension(self):
        """Check dimensions of configuration limit projection."""
        # The dimension of the barrier depends on the number of collision pairs
        self.assertEqual(
            SelfCollisionBarrier(
                n_collision_pairs=10,
                d_min=0.02,
            ).dim,
            10,
        )
        self.assertEqual(
            SelfCollisionBarrier(
                n_collision_pairs=len(
                    self.robot.collision_model.collisionPairs
                ),
                d_min=0.02,
            ).dim,
            len(self.robot.collision_model.collisionPairs),
        )

    def test_gains(self):
        """Check gains of configuration limit projection."""
        # One limit, scalar gain
        self.assertEqual(
            SelfCollisionBarrier(
                n_collision_pairs=10,
                d_min=0.02,
            ).gain.shape,
            (10,),
        )

    def test_jacobians(self):
        """Test that shapes of jacobians in all barriers are correct."""

        barrier = SelfCollisionBarrier(
            n_collision_pairs=len(self.robot.collision_model.collisionPairs),
            d_min=0.02,
        )
        J = barrier.compute_jacobian(self.configuration)
        self.assertEqual(J.ndim, 2)
        self.assertEqual(
            J.shape[0], len(self.robot.collision_model.collisionPairs)
        )

    def test_positive_when_in_safety_zone(self):
        """Check that the barrier is positive when in the safety zone."""

        barrier = SelfCollisionBarrier(
            n_collision_pairs=len(self.robot.collision_model.collisionPairs),
            d_min=0.02,
        )
        h = barrier.compute_barrier(self.configuration)
        self.assertTrue(np.all(h[self.non_colliding_objects_pair_id] > 0))

    def test_negative_when_out_of_safety_zone(self):
        """Test that the barrier is negative when out of the safety zone."""
        barrier = SelfCollisionBarrier(
            n_collision_pairs=len(self.robot.collision_model.collisionPairs),
            d_min=0.02,
        )
        # See https://github.com/stephane-caron/pink/pull/129 for an
        # illustration of this test configuration
        q_test = np.zeros(self.robot.model.nq)
        q_test[1] = 2.0
        q_test[3] = -2.5
        configuration = Configuration(
            self.robot.model,
            self.robot.data,
            q_test,
            collision_model=self.robot.collision_model,
            collision_data=self.robot.collision_data,
        )
        h = barrier.compute_barrier(configuration)
        self.assertTrue(np.min(h) < 0)  # there is a collision

    def test_closest_collision_pairs(self):
        """Test that the closest collision pairs are considered if number of
        collision pairs is less than actual number of collision pairs"""
        barrier = SelfCollisionBarrier(
            n_collision_pairs=5,
            d_min=0.02,
        )
        h = barrier.compute_barrier(self.configuration)
        distances = np.array(
            [
                self.configuration.collision_data.distanceResults[
                    k
                ].min_distance
                - 0.02
                for k in range(
                    len(self.configuration.collision_model.collisionPairs)
                )
            ]
        )
        for h_i in h:
            self.assertTrue(np.sum(distances < h_i) < barrier.dim)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""Test fixture for the frame task."""

import unittest

import numpy as np
import pinocchio as pin
from qpsolvers import solve_qp
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.exceptions import TargetNotSet, TaskDefinitionError
from pink.tasks import FrameTask


class TestFrameTask(unittest.TestCase):
    """Test consistency of the frame task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description(
            "jvrc_description", root_joint=pin.JointModelFreeFlyer()
        )
        frame_name = "ee_frame"
        joint_name = robot.model.names[-1]
        parent_joint = robot.model.getJointId(joint_name)
        parent_frame = robot.model.getFrameId(joint_name)
        placement = pin.SE3.Identity()
        robot.model.addFrame(
            pin.Frame(
                frame_name,
                parent_joint,
                parent_frame,
                placement,
                pin.FrameType.OP_FRAME,
            )
        )
        robot.data = pin.Data(robot.model)
        self.configuration = Configuration(robot.model, robot.data, robot.q0)

    def test_set_target_from_configuration(self):
        """Read target after setting it to a known value."""
        task = FrameTask("l_ankle", position_cost=1.0, orientation_cost=0.1)
        task.set_target_from_configuration(self.configuration)
        transform_ankle_to_world = (
            self.configuration.get_transform_frame_to_world("l_ankle")
        )
        self.assertTrue(
            np.allclose(
                transform_ankle_to_world, task.transform_target_to_world
            )
        )

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        earflap_task = FrameTask(
            "earflap", position_cost=1.0, orientation_cost=0.1
        )
        self.assertTrue("frame=" in repr(earflap_task))
        self.assertTrue("gain=" in repr(earflap_task))
        self.assertTrue("orientation_cost=" in repr(earflap_task))
        self.assertTrue("position_cost=" in repr(earflap_task))
        self.assertTrue("lm_damping=" in repr(earflap_task))

    def test_target_not_set(self):
        """Raise an exception when the target is not set."""
        task = FrameTask("l_ankle", position_cost=1.0, orientation_cost=0.1)
        with self.assertRaises(TargetNotSet):
            task.compute_error(self.configuration)
        with self.assertRaises(TargetNotSet):
            task.compute_jacobian(self.configuration)

    def test_target_set_properly(self):
        """Return target properly once it's set."""
        task = FrameTask("l_ankle", position_cost=1.0, orientation_cost=0.1)
        T = self.configuration.get_transform_frame_to_world("l_ankle")
        task.set_target(T)
        self.assertIsNotNone(task.transform_target_to_world)
        if task.transform_target_to_world is not None:  # help mypy
            self.assertTrue(
                np.allclose(
                    T.homogeneous,
                    task.transform_target_to_world.homogeneous,
                )
            )

    def test_target_is_a_copy(self):
        """Target is saved as a copy, not a reference to the original."""
        task = FrameTask("l_ankle", position_cost=1.0, orientation_cost=0.1)
        target = self.configuration.get_transform_frame_to_world("l_ankle")
        task.set_target(target)
        y = target.translation[1]
        target.translation[1] += 12.0
        if task.transform_target_to_world is not None:  # help mypy
            self.assertAlmostEqual(
                task.transform_target_to_world.translation[1], y
            )
            self.assertNotAlmostEqual(
                task.transform_target_to_world.translation[1],
                target.translation[1],
            )

    def test_zero_error_when_target_at_body(self):
        """Error is zero when the target and body are at the same location."""
        task = FrameTask("r_ankle", position_cost=1.0, orientation_cost=0.1)
        target = self.configuration.get_transform_frame_to_world("r_ankle")
        task.set_target(target)  # error == 0
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        J_ref = -self.configuration.get_frame_jacobian("r_ankle")  # at target
        self.assertLess(np.linalg.norm(e), 1e-10)
        self.assertTrue(np.allclose(J, J_ref))

    def test_unit_cost_qp_objective(self):
        """Unit cost means the QP objective is exactly (J^T J, -e^T J)."""
        task = FrameTask("r_wrist", position_cost=1.0, orientation_cost=0.1)
        transform_target_to_frame = pin.SE3(
            np.eye(3), np.array([0.0, 0.01, 0.0])
        )
        target = (
            self.configuration.get_transform_frame_to_world("r_wrist")
            * transform_target_to_frame
        )
        task.set_target(target)
        J = task.compute_jacobian(self.configuration)
        e = task.compute_error(self.configuration)
        task.set_position_cost(1.0)
        task.set_orientation_cost(1.0)
        task.lm_damping = 0.0
        H, c = task.compute_qp_objective(self.configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        self.assertTrue(np.allclose(e.T @ J, c))

    def test_zero_costs_same_as_disabling_lines(self):
        """Setting a position or orientation cost to zero.

        A zero yields the same QP objective as disabling the corresponding
        Jacobian and error coordinates.
        """
        task = FrameTask("l_wrist", position_cost=1.0, orientation_cost=0.1)
        transform_target_to_frame = pin.SE3(
            np.eye(3), np.array([0.1, 0.02, 0.01])
        )
        target = (
            self.configuration.get_transform_frame_to_world("l_wrist")
            * transform_target_to_frame
        )
        task.set_target(target)
        J = task.compute_jacobian(self.configuration)
        e = task.compute_error(self.configuration)
        qd = np.random.random(J.shape[1:])
        test_cases = {
            "position_only": (1.0, 0.0, slice(0, 3)),
            "orientation_only": (0.0, 1.0, slice(3, 6)),
            "position_0": ([1.0, 0.0, 0.0], 0.0, slice(0, 1)),
            "position_1": ([0.0, 1.0, 0.0], 0.0, slice(1, 2)),
            "position_2": ([0.0, 0.0, 1.0], 0.0, slice(2, 3)),
            "orientation_0": (0.0, [1.0, 0.0, 0.0], slice(3, 4)),
            "orientation_1": (0.0, [0.0, 1.0, 0.0], slice(4, 5)),
            "orientation_2": (0.0, [0.0, 0.0, 1.0], slice(5, 6)),
        }
        for case in test_cases.values():
            position_cost, orientation_cost, indexes = case
            task.set_position_cost(position_cost)
            task.set_orientation_cost(orientation_cost)
            task.lm_damping = 0.0
            H, c = task.compute_qp_objective(self.configuration)
            H_check = J[indexes].T @ J[indexes]
            c_check = e[indexes].T @ J[indexes]
            cost = qd.T @ H @ qd + c @ qd
            cost_check = qd.T @ H_check @ qd + c_check @ qd
            self.assertAlmostEqual(cost, cost_check)

    def test_lm_damping_has_no_effect_at_target(self):
        """Levenberg-Marquardt damping has no effect when the error is zero."""
        task = FrameTask("l_wrist", position_cost=1.0, orientation_cost=0.1)
        target = self.configuration.get_transform_frame_to_world("l_wrist")
        task.set_target(target)
        task.lm_damping = 1e-8
        H_1, c_1 = task.compute_qp_objective(self.configuration)
        task.lm_damping = 1e-4
        H_2, c_2 = task.compute_qp_objective(self.configuration)
        self.assertTrue(np.allclose(H_1, H_2))
        self.assertTrue(np.allclose(c_1, c_2))

    def test_lm_damping_has_effect_under_error(self):
        """Levenberg-Marquardt damping is indeed a damping.

        That is, unless the task is fulfilled, it reduces velocities.
        """
        task = FrameTask("r_wrist", position_cost=1.0, orientation_cost=0.1)
        transform_target_to_frame = pin.SE3(
            np.eye(3), np.array([0.0, 2.0, 0.0])
        )
        target = (
            self.configuration.get_transform_frame_to_world("r_wrist")
            * transform_target_to_frame
        )
        task.set_target(target)
        task.lm_damping = 1e-8
        H_1, c_1 = task.compute_qp_objective(self.configuration)
        task.lm_damping = 1e-4
        H_2, c_2 = task.compute_qp_objective(self.configuration)
        qd_1 = solve_qp(H_1, c_1, solver="daqp")  # H_1 p.s.d. (LM damping)
        qd_2 = solve_qp(H_2, c_2, solver="daqp")  # idem for H_2
        self.assertGreater(np.linalg.norm(qd_2 - qd_1), 1e-6)

    def test_task_on_user_added_op_frame(self):
        """Error is zero when the target and body are at the same location."""
        task = FrameTask("ee_frame", [1.0, 1.0, 1.0], [1.0, 1.0, 1.0])
        target = self.configuration.get_transform_frame_to_world("ee_frame")
        task.set_target(target)  # error == 0
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        J_ref = -self.configuration.get_frame_jacobian("ee_frame")  # at target
        self.assertLess(np.linalg.norm(e), 1e-10)
        self.assertTrue(np.allclose(J, J_ref))

    def test_inconsistent_cost(self):
        """Exception when the cost is not a vector any more."""
        task = FrameTask("ee_frame", [1.0, 1.0, 1.0], [1.0, 1.0, 1.0])
        task.cost = 42.0
        with self.assertRaises(TaskDefinitionError):
            task.set_position_cost(1.0)
        with self.assertRaises(TaskDefinitionError):
            task.set_orientation_cost(1.0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test configuration limit."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.limits import ConfigurationLimit


class TestConfigurationLimit(unittest.TestCase):
    """Test fixture for configuration limit."""

    def setUp(self):
        """Set test fixture up."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        model = robot.model
        self.data = robot.data
        self.limit = ConfigurationLimit(model)
        self.model = model
        self.robot = robot

    def test_dimensions(self):
        """Check dimensions of configuration limit projection."""
        for joint in self.limit.joints:
            self.assertGreaterEqual(joint.idx_q, 0)
            self.assertGreaterEqual(joint.idx_v, 0)
        nb = len(self.limit.joints)  # those are only revolute joints
        nv = self.model.nv
        self.assertEqual(len(self.limit.indices), nb)
        self.assertEqual(self.limit.projection_matrix.shape, (nb, nv))

    def test_model_with_no_limit(self):
        """Check that unbounded models don't fail."""
        empty_model = pin.Model()
        empty_bounded = ConfigurationLimit(empty_model)
        self.assertEqual(len(empty_bounded.indices), 0)

    def test_far_from_limit(self, tol=1e-10):
        """Limit has no effect when the configuration is far away.

        Args:
            tol: Numerical tolerance.

        When we are far away from configuration limits, the velocity limit is
        simply the configuration-agnostic one from the robot.

        Note:
            This test works with a commit-pinned description of Upkie. It will
            not work for any description, as a robot may have e.g. some
            velocity-unbounded joints.
        """
        robot = load_robot_description(
            "upkie_description",
            root_joint=pin.JointModelFreeFlyer(),
            commit="62f3ba24c2045b44faedb7c6c6167e74e157b49e",
        )
        dt = 1e-3  # [s]
        configuration = Configuration(robot.model, robot.data, robot.q0)
        limit = ConfigurationLimit(robot.model)
        G, h = limit.compute_qp_inequalities(robot.q0, dt=dt)
        v_lim = configuration.model.velocityLimit
        self.assertLess(np.max(+G @ v_lim * dt - h), -tol)
        self.assertLess(np.max(-G @ v_lim * dt - h), -tol)

    def test_configuration_limit_repulsion(self, tol=1e-10):
        """Velocities are scaled down when close to a configuration limit.

        Args:
            tol: Numerical tolerance.
        """
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        dt = 1e-3  # [s]
        configuration = Configuration(robot.model, robot.data, robot.q0)
        slack_vel = 5.5e-4  # [rad] / [s]

        # Clamp configuration limit to q +/- slack_vel * dt
        robot.model.lowerPositionLimit = configuration.integrate(
            -slack_vel * configuration.tangent.ones, dt
        )
        robot.model.upperPositionLimit = configuration.integrate(
            +slack_vel * configuration.tangent.ones, dt
        )

        limit = ConfigurationLimit(robot.model, config_limit_gain=0.5)
        G, h = limit.compute_qp_inequalities(configuration.q, dt)
        self.assertLess(np.max(h), slack_vel * dt + tol)
        self.assertGreater(np.min(h), -slack_vel * dt - tol)

    def test_gen2_description(self):
        robot = load_robot_description("gen2_description")
        limit = ConfigurationLimit(robot.model)
        self.assertEqual(tuple(limit.indices), (1, 2, 4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Build and solve the inverse kinematics problem."""

from typing import Iterable, Optional, Tuple

import numpy as np
import qpsolvers

from .barriers import Barrier
from .configuration import Configuration
from .exceptions import NoSolutionFound
from .limits import Limit
from .tasks import Task


def __compute_qp_objective(
    configuration: Configuration,
    tasks: Iterable[Task],
    damping: float,
    barriers: Optional[Iterable[Barrier]] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Compute the QP objective function.

    The Hessian matrix :math:`H` and linear vector :math:`c` define the QP
    objective function as:

    .. math::

        \frac{1}{2} \Delta q^T H \Delta q + c^T q

    The configuration displacement :math:`\Delta q` is the output of inverse
    kinematics (we divide it by :math:`\Delta t` to get a commanded velocity).

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks to fulfill at (weighted) best.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks. The same weight is applied to all joints, including if
            applicable the root joint (floating base, ...).
        dt: Integration timestep in [s].
        barriers: Collection of barriers.

    Returns:
        Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
        objective.
    """
    barriers = barriers if barriers is not None else []
    H = damping * configuration.tangent.eye
    c = configuration.tangent.zeros.copy()
    for task in tasks:
        H_task, c_task = task.compute_qp_objective(configuration)
        H += H_task
        c += c_task

    for barrier in barriers:
        H_cbf, c_cbf = barrier.compute_qp_objective(configuration)
        H += H_cbf
        c += c_cbf

    return (H, c)


def __compute_qp_inequalities(
    configuration: Configuration,
    limits: Optional[Iterable[Limit]],
    dt: float,
    barriers: Optional[Iterable[Barrier]] = None,
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    r"""Compute inequality constraints for the quadratic program.

    Args:
        configuration: Robot configuration to read kinematics from.
        limits: Collection of limits.
        dt: Integration timestep in [s].
        barriers: Collection of barriers.

    Returns:
        Pair :math:`(G, h)` of inequality matrix and vector representing the
        inequality :math:`G \Delta q \leq h`, or ``(None, None)`` if there is
        no inequality.

    Notes:
        We trim comparisons to infinity (equivalently: big floats) because some
        solvers don't support it. See for instance
        https://github.com/stephane-caron/pink/issues/10.
    """
    if limits is None:
        configuration_limit = configuration.model.configuration_limit
        velocity_limit = configuration.model.velocity_limit
        limits = (configuration_limit, velocity_limit)
    barriers = barriers if barriers is not None else []
    G_list = []
    h_list = []
    for limit in limits:
        matvec = limit.compute_qp_inequalities(configuration.q, dt)
        if matvec is not None:
            G_list.append(matvec[0])
            h_list.append(matvec[1])
    for barrier in barriers:
        G_barrier, h_barrier = barrier.compute_qp_inequalities(
            configuration, dt
        )
        G_list.append(G_barrier)
        h_list.append(h_barrier)
    if not G_list:
        return None, None
    return np.vstack(G_list), np.hstack(h_list)


def __compute_qp_equalities(
    configuration: Configuration,
    constraints: Optional[Iterable[Task]],
) -> Tuple[Optional[np.ndarray], Optional[np.ndarray]]:
    r"""Compute equality constraints for the quadratic program.

    Args:
        configuration: Robot configuration to read kinematics from.
        constraints: List of tasks to enforce via equality constraints.

    Returns:
        Pair :math:`(A, b)` of equality matrix and vector representing the
        equation :math:`A \Delta q = b`, or ``(None, None)`` if there is no
        equality constraint.
    """
    if not constraints:
        return None, None
    A_list = []
    b_list = []
    for task in constraints:
        jacobian = task.compute_jacobian(configuration)
        feedback = -task.gain * task.compute_error(configuration)
        A_list.append(jacobian)
        b_list.append(feedback)
    return np.vstack(A_list), np.hstack(b_list)


def build_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    damping: float = 1e-12,
    limits: Optional[Iterable[Limit]] = None,
    barriers: Optional[Iterable[Barrier]] = None,
    constraints: Optional[Iterable[Task]] = None,
) -> qpsolvers.Problem:
    r"""Build quadratic program from current configuration and tasks.

    This quadratic program is, in standard form:

    .. math::

        \begin{split}\begin{array}{ll}
            \underset{\Delta q}{\mbox{minimize}} &
                \frac{1}{2} {\Delta q}^T H {\Delta q} + c^T {\Delta q} \\
            \mbox{subject to}
                & G {\Delta q} \leq h
        \end{array}\end{split}

    where :math:`\Delta q` is a vector of joint displacements corresponding to
    the joint velocities :math:`v = {\Delta q} / {\mathrm{d}t}`.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks. The same weight is applied to all joints, including if
            applicable the root joint (floating base, ...).
        limits: Collection of limits to enforce. By default, consists of
            configuration and velocity limits (set to the empty list ``[]`` to
            disable limits).
        barriers: Collection of barriers.
        constraints: List of kinematic tasks to be enforced strictly, as hard
            equality constraints in the underlying QP rather than in its cost
            function.

    Returns:
        Quadratic program of the inverse kinematics problem.
    """
    P, q = __compute_qp_objective(configuration, tasks, damping, barriers)
    G, h = __compute_qp_inequalities(configuration, limits, dt, barriers)
    A, b = __compute_qp_equalities(configuration, constraints)
    problem = qpsolvers.Problem(P, q, G, h, A, b)
    return problem


def solve_ik(
    configuration: Configuration,
    tasks: Iterable[Task],
    dt: float,
    solver: str,
    damping: float = 1e-12,
    limits: Optional[Iterable[Limit]] = None,
    barriers: Optional[Iterable[Barrier]] = None,
    constraints: Optional[Iterable[Task]] = None,
    safety_break: bool = True,
    **kwargs,
) -> np.ndarray:
    r"""Compute a velocity tangent to the current robot configuration.

    The computed velocity satisfies at (weighted) best the set of kinematic
    tasks given in argument.

    Args:
        configuration: Robot configuration to read kinematics from.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        solver: Backend quadratic programming (QP) solver.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\mathrm{cost}]^2 / [\mathrm{tangent}]` where
            :math:`[\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks. The same weight is applied to all joints, including if
            applicable the root joint (floating base, ...).
        limits: Collection of limits to enforce. By default, consists of
            configuration and velocity limits (set to the empty list ``[]`` to
            disable limits).
        barriers: Collection of barrier functions.
        constraints: List of kinematic tasks to be enforced strictly, as hard
            equality constraints in the underlying QP rather than in its cost
            function.
        safety_break: If True, stop execution and raise an exception if
            the current configuration is outside limits. If False, print a
            warning and continue execution.
        kwargs: Keyword arguments to forward to the backend QP solver.

    Returns:
        Velocity :math:`v` in tangent space.

    Raises:
        NoSolutionFound: if the backend QP solver does not find a solution to
            the differential IK problem.
        NotWithinConfigurationLimits: if the current configuration is not
            within limits.

    Note:
        Our Tikhonov damping is isotropic despite tangent velocities not being
        homogeneous. If it helps we can add a tangent-space scaling to damp the
        floating base differently from joint angular velocities.
    """
    configuration.check_limits(safety_break=safety_break)
    problem = build_ik(
        configuration,
        tasks,
        dt,
        damping,
        limits,
        barriers,
        constraints,
    )
    result = qpsolvers.solve_problem(problem, solver=solver, **kwargs)
    Delta_q = result.x
    if not result.found or Delta_q is None:
        raise NoSolutionFound(problem, result)
    v: np.ndarray = Delta_q / dt
    return v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""G1 humanoid squat by regulating CoM"""

import numpy as np
import pinocchio as pin
import qpsolvers

import pink
from pink import solve_ik
from pink.tasks import ComTask, FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    ) from exc  # noqa: E501


if __name__ == "__main__":
    robot = load_robot_description(
        "g1_description", root_joint=pin.JointModelFreeFlyer()
    )

    # Initialize visualization
    viz = start_meshcat_visualizer(robot)

    q_ref = np.zeros(robot.nq)
    q_ref[2] = 0.72
    q_ref[6] = 1.0

    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    pelvis_orientation_task = FrameTask(
        "pelvis",
        position_cost=0.0,  # [cost] / [m]
        orientation_cost=10.0,  # [cost] / [rad]
    )

    com_task = ComTask(cost=200.0)
    com_task.set_target_from_configuration(configuration)

    posture_task = PostureTask(
        cost=1e-1,  # [cost] / [rad]
    )

    tasks = [pelvis_orientation_task, posture_task, com_task]

    for foot in ["right_ankle_roll_link", "left_ankle_roll_link"]:
        task = FrameTask(
            foot,
            position_cost=200.0,  # [cost] / [m]
            orientation_cost=10.0,  # [cost] / [rad]
        )
        tasks.append(task)

    for arm_points in ["right_palm_link", "left_palm_link"]:
        task = FrameTask(
            arm_points,
            position_cost=4.0,  # [cost] / [m]
            orientation_cost=0.0,  # [cost] / [rad]
        )
        tasks.append(task)

    for task in tasks:
        task.set_target_from_configuration(configuration)
        if isinstance(task, FrameTask):
            target = task.transform_target_to_world
            if task.frame in ["right_palm_link", "left_palm_link"]:
                target.translation += np.array([-0.1, 0.0, -0.2])
                task.set_target(target)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "osqp" in qpsolvers.available_solvers:
        solver = "osqp"

    rate = RateLimiter(frequency=200.0, warn=False)
    dt = rate.period
    t = 0.0  # [s]
    period = 2
    omega = 2 * np.pi / period

    while True:
        # Update CoM target
        Az = 0.05
        desired_com = np.zeros(3)
        desired_com[2] = 0.55 + Az * np.sin(omega * t)
        com_task.set_target(desired_com)

        velocity = solve_ik(
            configuration,
            tasks,
            dt,
            solver=solver,
            damping=0.01,
            safety_break=False,
        )
        configuration.integrate_inplace(velocity, dt)

        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Visualization helpers."""

import pinocchio as pin
from pinocchio import visualize


def start_meshcat_visualizer(
    robot: pin.RobotWrapper,
    open: bool = True,
) -> visualize.MeshcatVisualizer:
    """Open a MeshCat visualizer in a Web browser.

    Args:
        robot: Pinocchio robot wrapper with its model and data.
        open: If set (default), open MeshCat in a new Web browser tab.

    Returns:
        MeshCat visualizer.
    """
    visualizer = visualize.MeshcatVisualizer(
        robot.model, robot.collision_model, robot.visual_model
    )
    robot.setVisualizer(visualizer, init=False)
    visualizer.initViewer(open=open)
    visualizer.loadViewerModel()
    return visualizer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Load a custom URDF."""

import os

import pinocchio as pin

try:
    import yourdfpy
except ImportError:
    yourdfpy = None

if __name__ == "__main__":
    urdf_path = os.path.join(
        os.path.dirname(__file__),
        "robots",
        "double_pendulum.urdf",
    )

    robot = pin.RobotWrapper.BuildFromURDF(
        filename=urdf_path,
        package_dirs=["."],
        root_joint=None,
    )
    print(f"URDF description successfully loaded in {robot}")

    if yourdfpy is None:
        print("If you `pip install yourdfpy`, this example will display it.")
    else:  # yourdfpy is not None
        viz = yourdfpy.URDF.load(urdf_path)
        viz.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Upkie wheeled biped bending its knees."""

import meshcat_shapes
import numpy as np
import qpsolvers

import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.utils import custom_configuration_vector
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    )


if __name__ == "__main__":
    full_robot = load_robot_description("upkie_description", root_joint=None)
    robot = full_robot.buildReducedRobot(
        list_of_joints_to_lock=["left_wheel", "right_wheel"]
    )

    # Initialize visualization
    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["left_contact_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["right_contact_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["left_contact"], opacity=1.0)
    meshcat_shapes.frame(viewer["right_contact"], opacity=1.0)

    tasks = {
        "left_contact": FrameTask(
            "left_contact",
            position_cost=[0.1, 0.0, 0.1],  # [cost] / [m]
            orientation_cost=0.0,  # [cost] / [rad]
        ),
        "right_contact": FrameTask(
            "right_contact",
            position_cost=[0.1, 0.0, 0.1],  # [cost] / [m]
            orientation_cost=0.0,  # [cost] / [rad]
        ),
        "posture": PostureTask(
            cost=1e-3,  # [cost] / [rad]
        ),
    }

    q_ref = custom_configuration_vector(
        robot, left_hip=-0.2, left_knee=0.4, right_hip=0.2, right_knee=-0.4
    )
    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    for body, task in tasks.items():
        if type(task) is FrameTask:
            task.set_target_from_configuration(configuration)
    tasks["posture"].set_target(q_ref)
    viz.display(configuration.q)

    left_contact_target = tasks["left_contact"].transform_target_to_world
    right_contact_target = tasks["right_contact"].transform_target_to_world

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "daqp" in qpsolvers.available_solvers:
        solver = "daqp"

    rate = RateLimiter(frequency=200.0, warn=False)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        left_contact_target.translation[2] += 0.1 * np.sin(t) * dt
        right_contact_target.translation[2] += 0.1 * np.sin(t) * dt

        # Update visualization frames
        viewer["left_contact_target"].set_transform(left_contact_target.np)
        viewer["right_contact_target"].set_transform(right_contact_target.np)
        for body in ["left_contact", "right_contact"]:
            viewer[body].set_transform(
                configuration.get_transform_frame_to_world(body).np
            )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks.values(), dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Stéphane Caron, Simeon Nedelchev, Ivan Domrachev

"""Center of mass task implementation."""

from typing import Optional, Sequence, Union

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from ..exceptions import TargetNotSet, TaskDefinitionError
from .task import Task


class ComTask(Task):
    r"""Regulate the position of the center of mass.

    Attributes:
        target_com: Target position of the center of mass.

    Costs are designed so that errors with varying SI units, here position
    displacements, can be cast to homogeneous values. For example, if task
    "foo" has a position cost of 1.0 and task "bar" a position cost of 0.1,
    then a 1 [cm] error in task "foo" costs as much as a 10 [cm] error in task
    "bar".
    """

    target_com: Optional[np.ndarray]

    def __init__(
        self,
        cost: Union[float, Sequence[float]],
        lm_damping: float = 0.0,
        gain: float = 1.0,
    ) -> None:
        r"""Define a new center of mass task.

        Args:
            cost: Contribution of position errors to the normalized cost, in
                :math:`[\mathrm{cost}] / [\mathrm{m}]`. If this is a vector,
                the cost is anisotropic and each coordinate corresponds to an
                axis of the frame.
            lm_damping: Levenberg-Marquardt damping (see class attributes). The
                default value is conservatively low.
            gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
                filtering. Defaults to 1.0 (no filtering) for dead-beat
                control.
        """
        super().__init__(
            cost=np.ones(3),  # updated below
            gain=gain,
            lm_damping=lm_damping,
        )
        self.target_com = None
        self.set_cost(cost)

    def set_cost(
        self, cost: Union[float, Sequence[float], np.ndarray]
    ) -> None:
        r"""Set a new cost for all 3D CoM position coordinates.

        Args:
            cost: Contribution of position errors to the normalized cost, in
                :math:`[\mathrm{cost}] / [\mathrm{m}]`. If this is a vector,
                the cost is anisotropic and each coordinate corresponds to an
                axis of the frame.
        """
        if isinstance(cost, float):
            assert cost >= 0.0
        else:  # not isinstance(cost, float)
            assert all(cost >= 0.0 for cost in cost)
        if isinstance(self.cost, np.ndarray):
            self.cost[0:3] = cost
        else:  # self.cost is not a vector
            raise TaskDefinitionError(
                f"CoM task cost should be a vector, currently cost={self.cost}"
            )

    def set_target(
        self,
        target_com: np.ndarray,
    ) -> None:
        """Set the target CoM position in the world frame.

        Args:
            target_com: Target position of the center of mass in the world
                frame.
        """
        self.target_com = target_com.copy()

    def set_target_from_configuration(
        self, configuration: Configuration
    ) -> None:
        """Set the task target pose from a given robot configuration.

        Args:
            configuration: Robot configuration used to compute the target CoM.
        """
        q = configuration.q
        desired_com = pin.centerOfMass(
            configuration.model, configuration.data, q
        )
        self.set_target(desired_com)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the CoM task error.

        The error is the difference between the target CoM position and the
        current CoM position, both expressed in the world frame.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            CoM task error :math:`e(q)`.
        """
        if self.target_com is None:
            raise TargetNotSet("no target set for CoM")
        q = configuration.q
        actual_com = pin.centerOfMass(
            configuration.model, configuration.data, q
        )
        error = actual_com - self.target_com
        return error

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the CoM task Jacobian.

        The task Jacobian :math:`J(q) \in \mathbb{R}^{3 \times n_v}` is the
        derivative of the CoM position with respect to the configuration
        :math:`q`.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix :math:`J`.
        """
        if self.target_com is None:
            raise TargetNotSet("no target set for CoM")
        q = configuration.q
        J = pin.jacobianCenterOfMass(
            configuration.model, configuration.data, q
        )
        return J

    def __repr__(self):
        """Human-readable representation of the task."""
        cost = self.cost if isinstance(self.cost, float) else self.cost[0:3]
        return (
            "ComTask("
            f"target_com={self.target_com}, "
            f"cost={cost}, "
            f"gain={self.gain}, "
            f"lm_damping={self.lm_damping})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Low acceleration task."""

from typing import Optional

import numpy as np

from ..configuration import Configuration
from .posture_task import PostureTask


class LowAccelerationTask(PostureTask):
    r"""Minimize joint accelerations.

    The task is to minimize :math:`\| v \ominus v_{\mathrm{prev}} \|` where
    :math:`v_{\mathrm{prev}}` is the latest integrated velocity.

    Attributes:
        Delta_q_prev: Latest displacement of the robot.

    Note:
        This task does not dissipate energy and will thus tend to yield
        oscillations. Considering coupling it with a :class:`DampingTask`.
    """

    Delta_q_prev: Optional[np.ndarray]

    def __init__(self, cost: float) -> None:
        r"""Initialize task.

        Args:
            cost: joint angular acceleration cost, in
                :math:`[\mathrm{cost}] [\mathrm{s}]^2 / [\mathrm{rad}]`.
        """
        super().__init__(
            cost=cost,
            gain=1.0,  # this task hacks around the definition of error
            lm_damping=0.0,  # idem
        )
        self.Delta_q_prev = None

    def set_last_integration(self, v_prev: np.ndarray, dt) -> None:
        """Set the latest velocity and the duration it was applied for.

        The goal of the low-acceleration task is to minimize the difference
        between the new velocity and the previous one.

        Args:
            v_prev: Latest integrated velocity.
            dt: Integration timestep in [s].
        """
        self.Delta_q_prev = v_prev * dt

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the low-acceleration task error.

        The task error is opposite to the previous velocity, so that the
        overall task objective is :math:`\| v \ominus v_{\mathrm{prev}} \|^2`.

        Args:
            configuration: Robot configuration :math:`q`.

        Return:
            Low-acceleration task error
            :math:`e(q) = -\mathrm{d}t v_{\mathrm{prev}}`. This choice, along
            with :math:`\alpha = 1` and the Jacobian :math:`J` being the
            identity, ensures that :math:`J(q) \Delta q = -\alpha e(q)
            \Leftrightarrow v = \Delta q / \mathrm{d}t = v_{\mathrm{prev}}`.

        """
        Delta_q_prev: np.ndarray = (
            self.Delta_q_prev
            if self.Delta_q_prev is not None
            else np.zeros(configuration.model.nv)
        )
        return -Delta_q_prev

    def __repr__(self):
        """Human-readable representation of the task."""
        return f"LowAccelerationTask(cost={self.cost})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Kinematic tasks."""

from .com_task import ComTask
from .damping_task import DampingTask
from .frame_task import FrameTask
from .joint_coupling_task import JointCouplingTask
from .joint_velocity_task import JointVelocityTask
from .linear_holonomic_task import LinearHolonomicTask
from .low_acceleration_task import LowAccelerationTask
from .omniwheel_task import OmniwheelTask
from .posture_task import PostureTask
from .relative_frame_task import RelativeFrameTask
from .rolling_task import RollingTask
from .task import Task

__all__ = [
    "ComTask",
    "DampingTask",
    "FrameTask",
    "JointCouplingTask",
    "JointVelocityTask",
    "LinearHolonomicTask",
    "LowAccelerationTask",
    "OmniwheelTask",
    "PostureTask",
    "RelativeFrameTask",
    "RollingTask",
    "Task",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0

"""Unitree Z1 arm tracking a reference joint-velocity trajectory."""

import numpy as np
import qpsolvers

import pink
from pink import solve_ik
from pink.tasks import JointVelocityTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    )


if __name__ == "__main__":
    print(
        "In this example, the arm tracks a sinusoidal joint-velocity "
        "trajectory that is unfeasible at times.\nThe trajectory is only "
        "tracked while the robot stays within joint limits."
    )
    robot = load_robot_description("z1_description")
    viz = start_meshcat_visualizer(robot)
    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    viz.display(configuration.q)

    # Our only task in this example is a joint-velocity task
    joint_velocity_task = JointVelocityTask(cost=1.0)

    # Select a QP solver
    solver = qpsolvers.available_solvers[0]
    if "daqp" in qpsolvers.available_solvers:
        solver = "daqp"

    rate = RateLimiter(frequency=200.0, warn=False)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        target_velocity = 2.0 * np.sin(t) * np.ones(robot.nv)
        joint_velocity_task.set_target(target_velocity, dt)
        velocity = solve_ik(
            configuration,
            tasks=[joint_velocity_task],
            dt=dt,
            solver=solver,
        )
        configuration.integrate_inplace(velocity, dt)
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test velocity limit."""

import unittest

import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink.limits import VelocityLimit


class TestVelocityLimit(unittest.TestCase):
    """Test fixture for velocity limit."""

    def setUp(self):
        """Set test fixture up."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        model = robot.model
        self.limit = VelocityLimit(model)
        self.model = model

    def test_dimensions(self):
        """Check dimensions of velocity limit projection."""
        for joint in self.limit.joints:
            self.assertGreaterEqual(joint.idx_q, 0)
            self.assertGreaterEqual(joint.idx_v, 0)
        nb = len(self.limit.joints)  # those are only revolute joints
        nv = self.model.nv
        self.assertEqual(len(self.limit.indices), nb)
        self.assertEqual(self.limit.projection_matrix.shape, (nb, nv))

    def test_model_with_no_limit(self):
        """Check that unbounded models don't fail."""
        empty_model = pin.Model()
        empty_bounded = VelocityLimit(empty_model)
        self.assertEqual(len(empty_bounded.indices), 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Tests for position barrier limit."""

import unittest

import numpy as np
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.barriers import BodySphericalBarrier
from pink.exceptions import NegativeMinimumDistance


class TestBodySphericalBarrier(unittest.TestCase):
    """Testing position barrier limit."""

    def setUp(self):
        """Set test fixture up."""
        robot = load_robot_description("yumi_description")
        model = robot.model
        self.data = robot.data
        self.model = model
        self.robot = robot
        self.ees = ("yumi_link_7_l", "yumi_link_7_r")
        self.q0 = np.array(
            [
                0.045,
                -0.155,
                -0.394,
                -0.617,
                -0.939,
                -0.343,
                -1.216,
                0,
                0,
                -0.374,
                -0.249,
                0.562,
                -0.520,
                0.934,
                -0.337,
                1.400,
                0,
                0,
            ]
        )
        self.configuration = Configuration(self.model, self.data, self.q0)

    def test_negative_distance(self):
        """Raise an error when the robot body is not found."""
        with self.assertRaises(NegativeMinimumDistance):
            BodySphericalBarrier(self.ees, d_min=-1)

    def test_dimension(self):
        """Check dimensions of configuration limit projection."""
        self.assertEqual(BodySphericalBarrier(self.ees, d_min=0.2).dim, 1)

    def test_gains(self):
        """Check gains of configuration limit projection."""
        # One limit, scalar gain
        self.assertEqual(
            BodySphericalBarrier(self.ees, d_min=0.2).gain.shape, (1,)
        )

    def test_jacobians(self):
        """Test that shapes of jacobians in all barriers are correct."""

        barrier = BodySphericalBarrier(self.ees, d_min=0.2)
        J = barrier.compute_jacobian(self.configuration)
        self.assertEqual(J.ndim, 1)
        self.assertEqual(J.shape[0], self.robot.nv)

    def test_positive_when_in_safety_zone(self):
        """Check that the barrier is positive when in the safety zone."""

        barrier = BodySphericalBarrier(self.ees, d_min=0.2)
        h = barrier.compute_barrier(self.configuration)
        self.assertTrue(h[0] > 0)

    def test_negative_when_out_of_safety_zone(self):
        """Test that the barrier is negative when out of the safety zone."""

        barrier = BodySphericalBarrier(self.ees, d_min=0.3)
        h = barrier.compute_barrier(self.configuration)
        self.assertTrue(h[0] < 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2023 Inria

"""Subset of velocity-limited joints in a robot model."""

from typing import List, Optional, Tuple

import numpy as np
import pinocchio as pin

from .limit import Limit


class VelocityLimit(Limit):
    """Subset of velocity-limited joints in a robot model.

    Attributes:
        indices: Tangent indices corresponding to velocity-limited joints.
        joints: List of velocity-limited joints.
        model: Robot model.
        projection_matrix: Projection from tangent space to subspace with
            velocity-limited joints.
    """

    indices: np.ndarray
    joints: list
    model: pin.Model
    projection_matrix: Optional[np.ndarray]

    def __init__(self, model: pin.Model):
        """Initialize bounded tangent of a model.

        Args:
            model: robot model.
        """
        has_velocity_limit = np.logical_and(
            model.velocityLimit < 1e20,
            model.velocityLimit > 1e-10,
        )

        joints = [
            joint
            for joint in model.joints
            if joint.idx_v >= 0
            and has_velocity_limit[
                slice(joint.idx_v, joint.idx_v + joint.nv)
            ].all()
        ]

        index_list: List[int] = []
        for joint in joints:
            index_list.extend(range(joint.idx_v, joint.idx_v + joint.nv))
        indices = np.array(index_list)
        indices.setflags(write=False)

        dim = len(indices)
        projection_matrix = np.eye(model.nv)[indices] if dim > 0 else None

        self.indices = indices
        self.joints = joints
        self.model = model
        self.projection_matrix = projection_matrix

    def compute_qp_inequalities(
        self,
        q: np.ndarray,
        dt: float,
    ) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        r"""Compute inequalities for velocity limits.

        Those limits are defined by:

        .. math::

            -\mathrm{d}t v_{max} \leq \Delta q \leq \mathrm{d}t v_{max}

        where :math:`v_{max} \in {\cal T}` is the robot's velocity limit
        vector and :math:`\Delta q \in T_q({\cal C})` is the displacement
        computed by the inverse kinematics.

        Args:
            q: Robot configuration.
            dt: Integration timestep in [s].

        Returns:
            Pair :math:`(G, h)` representing the inequality constraint as
            :math:`G \Delta q \leq h`, or ``None`` if there is no limit.
        """
        if self.projection_matrix is None:  # no joint (thus checked for mypy)
            return None

        v_max = self.model.velocityLimit[self.indices]
        G = np.vstack([self.projection_matrix, -self.projection_matrix])
        h = np.hstack([dt * v_max, dt * v_max])
        return G, h



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""SigmaBan humanoid standing on two feet."""

import numpy as np
import pinocchio as pin
import qpsolvers

import meshcat_shapes
import pink
from pink import solve_ik
from pink.tasks import FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    )


if __name__ == "__main__":
    robot = load_robot_description(
        "sigmaban_description", root_joint=pin.JointModelFreeFlyer()
    )

    # Initialize visualization
    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["left_foot_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["right_foot_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["torso_target"], opacity=0.5)

    configuration = pink.Configuration(robot.model, robot.data, robot.q0)
    viz.display(configuration.q)

    left_foot_task = FrameTask(
        "left_foot_tip",
        position_cost=1.0,
        orientation_cost=1.0,
    )
    torso_task = FrameTask(
        "torso",
        position_cost=1.0,
        orientation_cost=1.0,
    )
    right_foot_task = FrameTask(
        "right_foot_tip",
        position_cost=1.0,
        orientation_cost=1.0,
    )
    posture_task = PostureTask(
        cost=1e-2,  # [cost] / [rad]
    )
    tasks = [left_foot_task, torso_task, right_foot_task, posture_task]

    torso_pose = configuration.get_transform_frame_to_world("torso").copy()
    # torso_pose.translation[0] += 0.05
    torso_task.set_target(torso_pose)
    posture_task.set_target_from_configuration(configuration)

    transform_left_foot_tip_target_to_init = pin.SE3(
        np.eye(3), np.array([0.0, 0.03, 0.0])
    )
    transform_right_foot_tip_target_to_init = pin.SE3(
        np.eye(3), np.array([0.0, -0.03, 0.0])
    )

    left_foot_task.set_target(
        configuration.get_transform_frame_to_world("left_foot_tip")
        * transform_left_foot_tip_target_to_init
    )
    right_foot_task.set_target(
        configuration.get_transform_frame_to_world("right_foot_tip")
        * transform_right_foot_tip_target_to_init
    )
    torso_task.set_target(configuration.get_transform_frame_to_world("torso"))

    # Display targets
    viewer["left_foot_target"].set_transform(
        left_foot_task.transform_target_to_world.np
    )
    viewer["right_foot_target"].set_transform(
        right_foot_task.transform_target_to_world.np
    )
    viewer["torso_target"].set_transform(
        torso_task.transform_target_to_world.np
    )

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "daqp" in qpsolvers.available_solvers:
        solver = "daqp"

    rate = RateLimiter(frequency=200.0, warn=False)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Tests that should pass for all barriers."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.barriers import PositionBarrier


class TestBarrier(unittest.TestCase):
    """Tests that should pass for all barriers."""

    def setUp(self):
        """Set up test fixtures."""

        self.robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        self.conf = Configuration(
            self.robot.model,
            self.robot.data,
            q=np.zeros(self.robot.nq),
        )
        self.dt = 1e-3  # [s]

    def test_diff_form_dimensions(self):
        """Velocity barrier dimension is the number of bounded joints."""
        for barrier in [
            PositionBarrier("left_hip", p_min=np.zeros(3), p_max=np.zeros(3))
        ]:
            H, c = barrier.compute_qp_objective(self.conf)
            G, h = barrier.compute_qp_inequalities(self.conf, self.dt)
            self.assertEqual(H.shape[0], self.robot.nv)
            self.assertEqual(H.shape[1], self.robot.nv)
            self.assertEqual(c.shape[0], self.robot.nv)
            self.assertEqual(G.shape[0], barrier.dim)
            self.assertEqual(G.shape[1], self.robot.nv)

    def test_barrier_value_dimension(self):
        """Test tha shape of value in all barriers is correct."""

        for barrier in [
            PositionBarrier("left_hip", p_min=np.zeros(3), p_max=np.zeros(3))
        ]:
            v = barrier.compute_barrier(self.conf)
            self.assertEqual(v.shape[0], barrier.dim)

    def test_barrier_jacobians_dimension(self):
        """Test that shapes of jacobians in all barriers are correct."""

        for barrier in [
            PositionBarrier("left_hip", p_min=np.zeros(3), p_max=np.zeros(3))
        ]:
            J = barrier.compute_jacobian(self.conf)
            self.assertEqual(J.shape[0], barrier.dim)
            self.assertEqual(J.shape[1], self.robot.nv)

    def test_barrier_without_penalty_weight(self):
        """Test that objective is zero if no penalty weight is provided."""
        for barrier in [
            PositionBarrier("left_hip", p_min=np.zeros(3), p_max=np.zeros(3))
        ]:
            H, c = barrier.compute_qp_objective(self.conf)
            self.assertTrue(np.allclose(H, 0))
            self.assertTrue(np.allclose(c, 0))

    def test_barrier_penalty_weight(self):
        """Test that objective is non-zeros, if penalty weight is provided."""
        for barrier in [
            PositionBarrier(
                "left_hip",
                p_min=np.zeros(3),
                p_max=np.zeros(3),
                safe_displacement_gain=1.0,
            )
        ]:
            H, c = barrier.compute_qp_objective(self.conf)
            self.assertFalse(np.allclose(H, 0))
            if np.any(barrier.compute_safe_displacement(self.conf) != 0):
                self.assertFalse(np.allclose(c, 0))

    def test_task_repr(self):
        """Test task string representation."""
        for limit in [
            PositionBarrier(
                "universe", safe_displacement_gain=0.0, p_min=np.zeros(3)
            )
        ]:
            self.assertTrue("gain=" in repr(limit))
            self.assertTrue("safe_displacement=" in repr(limit))
            self.assertTrue("safe_displacement_gain" in repr(limit))
            self.assertTrue("dim" in repr(limit))

    def test_cached(self):
        """Test that cached results are reused."""

        barrier = PositionBarrier(
            "left_hip", p_min=np.zeros(3), p_max=np.zeros(3)
        )
        barrier.compute_qp_objective(self.conf)
        # Check that cache is initially triggered
        self.assertIsNotNone(barrier._Barrier__q_cache)

        # Check that cache is resetted after update
        self.conf.update(self.conf.q + 1.0)
        barrier.compute_qp_inequalities(self.conf)
        self.assertTrue(np.allclose(barrier._Barrier__q_cache, self.conf.q))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Yumi two-armed manipulator with definition of custom frame and
collision avoidance w.r.t. those frames."""

import numpy as np
import pinocchio as pin
import qpsolvers

import meshcat_shapes
import pink
from pink import solve_ik
from pink.barriers import BodySphericalBarrier
from pink.tasks import FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    ) from exc  # noqa: E501


if __name__ == "__main__":
    # Load the robot and define the custom frames
    robot = load_robot_description("yumi_description", root_joint=None)

    # Left arm frame
    l_frame_placement = pin.SE3()
    l_frame_placement.translation = np.array([0.0, 0.0, 0.05])
    l_frame_placement.rotation = np.eye(3)

    l_frame = pin.Frame(
        "yumi_barrier_l",
        robot.model.getJointId("yumi_joint_6_l"),
        robot.model.getFrameId("yumi_link_7_l"),
        l_frame_placement,
        pin.FrameType.OP_FRAME,
    )

    robot.model.addFrame(l_frame)

    # Right arm frame
    r_frame_placement = pin.SE3()
    r_frame_placement.translation = np.array([0.0, 0.0, 0.05])
    r_frame_placement.rotation = np.eye(3)

    r_frame = pin.Frame(
        "yumi_barrier_r",
        robot.model.getJointId("yumi_joint_6_r"),
        robot.model.getFrameId("yumi_link_7_r"),
        r_frame_placement,
        pin.FrameType.OP_FRAME,
    )

    robot.model.addFrame(r_frame)
    # Redefining robot data to take into account the new frames
    robot.data = pin.Data(robot.model)

    viz = start_meshcat_visualizer(robot)

    # Pink tasks
    left_end_effector_task = FrameTask(
        "yumi_link_7_l",
        position_cost=50.0,  # [cost] / [m]
        orientation_cost=10.0,  # [cost] / [rad]
    )
    right_end_effector_task = FrameTask(
        "yumi_link_7_r",
        position_cost=50.0,  # [cost] / [m]
        orientation_cost=10.0,  # [cost] / [rad]
    )

    # Pink barriers
    ee_barrier = BodySphericalBarrier(
        ("yumi_barrier_l", "yumi_barrier_r"),
        d_min=0.25,
        gain=100.0,
        safe_displacement_gain=1.0,
    )

    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )

    cbf_list = [ee_barrier]
    tasks = [left_end_effector_task, right_end_effector_task, posture_task]

    q_ref = np.array(
        [
            0.045,
            -0.155,
            -0.394,
            -0.617,
            -0.939,
            -0.343,
            -1.216,
            0,
            0,
            -0.374,
            -0.249,
            0.562,
            -0.520,
            0.934,
            -0.337,
            1.400,
            0,
            0,
        ]
    )
    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    for task in tasks:
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    viewer = viz.viewer
    meshcat_shapes.frame(viewer["left_end_effector"], opacity=1.0)
    meshcat_shapes.sphere(
        viewer["left_ee_barrier"],
        opacity=0.4,
        color=0xFF0000,
        radius=0.125,
    )
    meshcat_shapes.sphere(
        viewer["right_ee_barrier"],
        opacity=0.4,
        color=0x00FF00,
        radius=0.125,
    )
    meshcat_shapes.frame(viewer["right_end_effector"], opacity=1.0)
    meshcat_shapes.frame(viewer["left_end_effector_target"], opacity=1.0)
    meshcat_shapes.frame(viewer["right_end_effector_target"], opacity=1.0)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "osqp" in qpsolvers.available_solvers:
        solver = "osqp"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    l_y_des = np.array([0.392, 0.392, 0.6])
    r_y_des = np.array([0.392, 0.392, 0.6])
    l_dy_des = np.zeros(3)
    r_dy_des = np.zeros(3)

    while True:
        # Calculate desired trajectory
        A = 0.1
        B = 0.1
        # z -- 0.4 - 0.8
        l_y_des[:] = (
            0.6,
            0.1 + B * np.sin(t),
            0.6 + A * np.sin(t),
        )
        r_y_des[:] = (
            0.6,
            -0.1 - B * np.sin(t),
            0.6 + A * np.sin(t),
        )
        l_dy_des[:] = 0, B * np.cos(t), A * np.cos(t)
        r_dy_des[:] = 0, -B * np.cos(t), A * np.cos(t)

        left_end_effector_task.transform_target_to_world.translation = l_y_des
        right_end_effector_task.transform_target_to_world.translation = r_y_des

        # Update visualization frames
        viewer["left_end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                left_end_effector_task.frame
            ).np
        )
        viewer["right_end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                right_end_effector_task.frame
            ).np
        )
        viewer["left_end_effector_target"].set_transform(
            left_end_effector_task.transform_target_to_world.np
        )
        viewer["right_end_effector_target"].set_transform(
            right_end_effector_task.transform_target_to_world.np
        )

        lb = configuration.get_transform_frame_to_world("yumi_barrier_l")
        rb = configuration.get_transform_frame_to_world("yumi_barrier_r")

        viewer["left_ee_barrier"].set_transform(lb.np)
        viewer["right_ee_barrier"].set_transform(rb.np)

        velocity = solve_ik(
            configuration,
            tasks,
            dt,
            solver=solver,
            barriers=cbf_list,
        )
        configuration.integrate_inplace(velocity, dt)
        dist_ee = (
            np.linalg.norm(
                configuration.get_transform_frame_to_world(
                    "yumi_barrier_l"
                ).translation
                - configuration.get_transform_frame_to_world(
                    "yumi_barrier_r"
                ).translation
            )
            * 100
        )
        print(f"Distance between end effectors: {dist_ee :0.1f}cm")
        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""UR3 arm tracking a target, first without then with velocity smoothing."""

import matplotlib.pyplot as plt
import numpy as np
import qpsolvers

import meshcat_shapes
import pink
from pink import solve_ik
from pink.limits import AccelerationLimit
from pink.tasks import DampingTask, FrameTask, PostureTask
from pink.utils import custom_configuration_vector
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Examples need the 'robot_descriptions' package "
        "(available from conda-forge or PyPI)"
    )

NB_STEPS = 3000  # number of steps to run the example for

if __name__ == "__main__":
    robot = load_robot_description("ur3_description", root_joint=None)

    viz = start_meshcat_visualizer(robot)
    viewer = viz.viewer
    meshcat_shapes.frame(viewer["end_effector_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["end_effector"], opacity=1.0)

    # Define inverse kinematics tasks and limits
    end_effector_task = FrameTask(
        "ee_link",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
        lm_damping=1.0,  # tuned for this setup
    )
    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )
    damping_task = DampingTask(
        cost=1e-1,  # [cost] * [s] / [rad]
    )
    acceleration_limit = AccelerationLimit(
        robot.model,
        np.full(
            robot.model.nv,
            20.0,  # [rad] / [s]^2
        ),
    )

    # Initial configuration and task setup
    q_ref = custom_configuration_vector(
        robot,
        shoulder_lift_joint=1.0,
        shoulder_pan_joint=1.0,
        elbow_joint=1.0,
    )
    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    end_effector_task.set_target_from_configuration(configuration)
    posture_task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "daqp" in qpsolvers.available_solvers:
        solver = "daqp"

    rate = RateLimiter(frequency=200.0, warn=False)
    dt = rate.period
    configurations, velocities, times = [], [], []
    t = 0.0  # [s]
    max_ever = -1.0
    for step in range(NB_STEPS):
        # Update task targets
        end_effector_target = end_effector_task.transform_target_to_world
        end_effector_target.translation[1] = 0.5 + 0.1 * np.sin(2.0 * t)
        end_effector_target.translation[2] = 0.2

        # Update visualization frames
        viewer["end_effector_target"].set_transform(end_effector_target.np)
        viewer["end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                end_effector_task.frame
            ).np
        )

        if step < NB_STEPS // 2:
            # First half: no velocity smoothing
            end_effector_task.gain = 1.0
            tasks = (end_effector_task, posture_task)
            limits = (
                configuration.model.configuration_limit,
                configuration.model.velocity_limit,
            )
        else:  # step >= NB_STEPS // 2
            # Second half: velocity smoothing by:
            # 1. Reducing the task gain
            # 2. Switching from a posture to a damping task
            # 3. Adding an acceleration limit
            end_effector_task.gain = 0.4
            tasks = (end_effector_task, damping_task)
            limits = (
                configuration.model.configuration_limit,
                configuration.model.velocity_limit,
                acceleration_limit,
            )

        # Compute velocity and integrate it into next configuration
        velocity = solve_ik(
            configuration, tasks, dt, solver=solver, limits=limits
        )
        Delta_q = velocity * dt
        configuration.integrate_inplace(velocity, dt)
        acceleration_limit.set_last_integration(velocity, dt)

        # Append plotting data to lists
        configurations.append(configuration.q)
        times.append(t)
        velocities.append(velocity)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt

    # Plot velocities and configurations over time
    velocities = np.array(velocities)
    configurations = np.array(configurations)
    plt.figure()
    plt.plot(times, velocities)
    plt.plot(
        [times[NB_STEPS // 2]] * 2,
        [velocities.min(), velocities.max()],
        "k--",
    )
    plt.title("Joint velocities (posture then damping task)")
    plt.xlabel("Time [s]")
    plt.ylabel("Joint velocity [rad/s]")
    plt.figure()
    plt.plot(times, configurations)
    plt.plot(
        [times[NB_STEPS // 2]] * 2,
        [configurations.min(), configurations.max()],
        "k--",
    )
    plt.title("Configuration (posture then damping task)")
    plt.xlabel("Time [s]")
    plt.ylabel("Joint position [rad]")
    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test fixture for the posture task."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.exceptions import TargetNotSet
from pink.tasks import PostureTask


class TestPostureTask(unittest.TestCase):
    """Test consistency of the posture task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description(
            "upkie_description", root_joint=pin.JointModelFreeFlyer()
        )
        self.configuration = Configuration(robot.model, robot.data, robot.q0)

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        task = PostureTask(cost=1.0)
        self.assertTrue("cost=" in repr(task))
        self.assertTrue("gain=" in repr(task))

    def test_target_not_set(self):
        """Raise an exception when the target is not set."""
        task = PostureTask(cost=1.0)
        with self.assertRaises(TargetNotSet):
            task.compute_error(self.configuration)

    def test_set_target_from_configuration(self):
        """Check that target is set from its configuration value."""
        task = PostureTask(cost=1.0)
        task.set_target_from_configuration(self.configuration)
        self.assertTrue(np.allclose(self.configuration.q, task.target_q))

    def test_target_set_properly(self):
        """Return target properly once it's set."""
        task = PostureTask(cost=1.0)
        task.set_target(self.configuration.q)
        self.assertIsNotNone(task.target_q)
        if task.target_q is not None:  # help mypy
            self.assertTrue(np.allclose(self.configuration.q, task.target_q))

    def test_target_is_a_copy(self):
        """Target is saved as a copy, not a reference to the original."""
        task = PostureTask(cost=1.0)
        q = self.configuration.q
        with self.assertRaises(ValueError):
            q[2] += 12.0  # assignment destination is read-only
        task.set_target(q)
        self.assertIsNotNone(task.target_q)

    def test_zero_error_when_target_at_body(self):
        """Error is zero when the target and body are at the same location."""
        task = PostureTask(cost=1.0)
        task.set_target(self.configuration.q)  # error == 0
        e = task.compute_error(self.configuration)
        self.assertLess(np.linalg.norm(e), 1e-10)

    def test_unit_cost_qp_objective(self):
        """A unit cost vector means the QP objective is (J^T J, -e^T J)."""
        task = PostureTask(cost=1.0)
        task.set_target(self.configuration.q)
        q_new = self.configuration.q.copy()
        q_new[1] += 1.0
        q_new[3] += 1.0
        q_new[5] += 1.0
        self.configuration.q = q_new
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        self.assertTrue(np.allclose(-e.T @ J, c))

    def test_zero_cost_same_as_disabling_task(self):
        """The task has no effect when its cost is zero."""
        task = PostureTask(cost=0.0)
        q = self.configuration.q
        task.set_target(q)
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        qd = np.random.random(J.shape[1:])
        cost = qd.T @ H @ qd + c @ qd
        self.assertAlmostEqual(cost, 0.0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2025 Inria

"""Test fixture for the omniwheel task."""

import unittest

from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.tasks import OmniwheelTask


class TestOmniwheelTask(unittest.TestCase):
    """Test consistency of the omniwheel task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        robot = load_robot_description("upkie_description")
        left_wheel_task = OmniwheelTask(
            "left_wheel_center",
            floor_frame="universe",
            wheel_radius=0.1,
            cost=1.0,
        )
        self.configuration = Configuration(robot.model, robot.data, robot.q0)
        self.left_wheel_task = left_wheel_task

    def test_compute_error(self):
        task = self.left_wheel_task
        error_in_rim = task.compute_error(self.configuration)
        self.assertAlmostEqual(error_in_rim[0], 0.0)

    def test_compute_jacobian(self):
        task = self.left_wheel_task
        jacobian_hub_in_rim = task.compute_jacobian(self.configuration)
        self.assertEqual(
            jacobian_hub_in_rim.shape, (2, self.configuration.model.nv)
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Rolling task implementation."""

from typing import Optional, Sequence, Union

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from .task import Task


class RollingTask(Task):
    r"""Roll without slipping on a plane.

    Frames used in this task are:

    - Floor: inertial frame defining the contact plane the wheel rolls on.
    - Hub: frame attached to the hub of the wheel in the robot model.
    - Rim: its origin is the projection of the hub origin onto the floor
      plane. Its orientation is the same as the floor frame.

    Attributes:
        hub_frame: Name of a frame attached to the hub of the wheel in the
            robot model.
        floor_frame: Name of the inertial frame whose xy-plane defines the
            contact surface the wheel is rolling onto.
        wheel_radius: Radius of the wheel, i.e. distance in meters from the hub
            to the nearest point on the rim.

    Note:
        See also the similar `wheel task in the PlaCo C++ library
        <https://github.com/Rhoban/placo/blob/9e65abadff45071ab11274eb74770d71e10b7ca8/docs/kinematics/wheel_task.rst>`__.
    """

    hub_frame: str
    floor_frame: str
    wheel_radius: float

    def __init__(
        self,
        hub_frame: str,
        floor_frame: str,
        wheel_radius: float,
        cost: Optional[Union[float, Sequence[float], np.ndarray]],
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        r"""Define a new rolling task.

        Args:
            hub_frame: Name of a frame attached to the hub of the wheel in the
                robot model.
            floor_frame: Name of the inertial frame whose xy-plane defines the
                contact surface the wheel is rolling onto.
            wheel_radius: Radius of the wheel, i.e. distance in meters from the
                hub to the nearest point on the rim.
            cost: scalar or 3D cost vector,
                in :math:`[\mathrm{cost}] / [\mathrm{m}]`.
            lm_damping: Levenberg-Marquardt damping (see class attributes). The
                default value is conservatively low.
            gain: Task gain :math:`\alpha \in [0, 1]` for additional low-pass
                filtering. Defaults to 1.0 (no filtering) for dead-beat
                control.
        """
        super().__init__(
            cost=cost,
            gain=gain,
            lm_damping=lm_damping,
        )
        self.floor_frame = floor_frame
        self.hub_frame = hub_frame
        self.wheel_radius = wheel_radius

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the rolling task error.

        The error is a 3D vector :math:`{}_R e(q)` of the rim frame :math:`R`:

        .. math::

            {}_R e(q) := \begin{bmatrix}
                0 \\
                0 \\
                z_{\mathit{hub}} - \rho
                \end{bmatrix}

        The error is zero along the x- and y-axis of the rim frame to roll
        without slipping (*i.e.*, the velocity of the contact point is zero in
        the floor plane). The error along the z-axis of the rim frame is chosen
        so that the task keeps the wheel hub at distance :math:`\rho` (wheel
        radius) from the floor plane.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task error :math:`{}_R e(q) \in \mathbb{R}^3`, a translation
            expressed in the rim frame :math:`R`.
        """
        transform_hub_to_floor = configuration.get_transform(
            source=self.hub_frame, dest=self.floor_frame
        )
        position_hub_in_floor = transform_hub_to_floor.translation
        error_in_rim = np.zeros(3)
        error_in_rim[2] = position_hub_in_floor[2] - self.wheel_radius
        return error_in_rim

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the rolling task Jacobian.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Translation Jacobian matrix :math:`{}_R J_{WR}`, expressed locally
            in the rim frame :math:`R`.
        """
        transform_hub_to_floor = configuration.get_transform(
            source=self.hub_frame, dest=self.floor_frame
        )
        position_hub_in_floor = transform_hub_to_floor.translation
        position_rim_in_floor = position_hub_in_floor + np.array(
            [0.0, 0.0, -self.wheel_radius]
        )
        transform_rim_to_floor = pin.SE3(
            rotation=np.eye(3),
            translation=position_rim_in_floor,
        )
        transform_hub_to_rim = transform_rim_to_floor.actInv(
            transform_hub_to_floor
        )
        jacobian_hub_in_hub = configuration.get_frame_jacobian(self.hub_frame)
        jacobian_hub_in_rim = transform_hub_to_rim.action @ jacobian_hub_in_hub
        return jacobian_hub_in_rim[:3]  # translation

    def __repr__(self):
        """Human-readable representation of the task."""
        return (
            "RollingTask("
            f"hub_frame={self.hub_frame}, "
            f"floor_frame={self.floor_frame}, "
            f"wheel_radius={self.wheel_radius}, "
            f"cost={self.cost}, "
            f"gain={self.gain}, "
            f"lm_damping={self.lm_damping})"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Self Collision Avoidance Barrier with hpp-fcl."""

from typing import Union

import numpy as np
import pinocchio as pin

from ..configuration import Configuration
from ..exceptions import InvalidCollisionPairs, NegativeMinimumDistance
from .barrier import Barrier


class SelfCollisionBarrier(Barrier):
    r"""Self Collision Avoidance Barrier.

    This class defines a barrier function based on the smooth convex collision
    geometry. using hpp-fcl through Pinocchio. Note that for non-smooth
    collision geometries behaviour is undefined.

    The barrier function is defined as:

    .. math::

        h(q) = \begin{bmatrix} \ldots \\ d(p^1_i, p^2_i) - d_{min}^2 \\
            \ldots \end{bmatrix} \quad \forall i \in 0 \ldots N

    where :math:`N` is number of collision pairs, :math:`p^k_i` is the
    :math:`k-th` body in :math:`i`-th collision pair, :math:`d(p^1_i, p^2_i)`
    is the distance between collision bodies in the pair, and :math:`d_{min}`
    is minimal distance between any collision bodies.

    Note:
        The number of evaluated collision pairs might not be equal to the
        number of all collision pairs. If the former is lower, then only the
        closest collision pairs will be considered.

    Attributes:
        d_min: Minimum distance between collision pairs.
    """

    d_min: float

    def __init__(
        self,
        n_collision_pairs: int,
        gain: Union[float, np.ndarray] = 1.0,
        safe_displacement_gain: float = 1.0,
        d_min: float = 0.02,
    ):
        """Initialization of the SelfCollisionBarrier.

        Args:
            n_collision_pairs: Number of collision pairs.
                Note that the number of collision pairs doesn't have to be
                equal to the total number of collision pairs in the model. If
                it is less, than only the closest collision pairs will be used.
            gain: Barrier gain. Defaults to 1.0.
            safe_displacement_gain: gain for the safe backup displacement.
                cost term. Defaults to 1.0.
            d_min: Minimum distance between any collision pairs.
                Defaults to 0.02.
        """
        if d_min < 0.0:
            raise NegativeMinimumDistance(
                "The minimum distance threshold must be non-negative."
            )
        if n_collision_pairs < 0:
            raise InvalidCollisionPairs(
                "The number of collision pairs must be non-negative."
            )

        super().__init__(
            dim=n_collision_pairs,
            gain=gain,
            safe_displacement_gain=safe_displacement_gain,
        )

        self.d_min = d_min

    def compute_barrier(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the value of the barrier function.

        The barrier function is computed as the vector of lowest distances
        between collision pairs. It is defined as:

        .. math::

            h(q) = \begin{bmatrix} \ldots \\ d(p^1_i, p^2_i) - d_{min}^2 \\
                \ldots \end{bmatrix} \quad \forall i \in 1 \ldots N

        where :math:`N` is number of collision pairs, :math:`p^k_i` is the
        :math:`k-th` body in :math:`i`-th collision pair, :math:`d(p^1_i,
        p^2_i)` is the distance between collision bodies in the pair, and
        :math:`d_{min}` is the minimal distance between any collision bodies.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Value of the barrier function
                 :math:`h(q)`.
        """
        if len(configuration.collision_model.collisionPairs) < self.dim:
            raise InvalidCollisionPairs(
                "The number of collision pairs "
                f"({len(configuration.collision_model.collisionPairs)}) "
                f"is less than the barrier dimension ({self.dim})."
            )
        distances = np.array(
            [
                configuration.collision_data.distanceResults[k].min_distance
                - self.d_min
                for k in range(
                    len(configuration.collision_model.collisionPairs)
                )
            ]
        )
        closest_pairs_idxs = np.argpartition(-distances, -self.dim)[
            -self.dim :
        ]

        return distances[closest_pairs_idxs]

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the Jacobian matrix of the barrier function.

        The Jacobian matrix could be represented as stacked gradients of each
        collision pair distance function w.r.t. joints. They are computed based
        on joints Jacobians and normal surface vector at nearest distance
        points n.

        The gradient, *a.k.a.* the i-th row in the Jacobian matrix, is given
        by:

        .. math::

            J_i = n_1^T J^1_p + (r_1 \times n_1)^T J^1_w +
                n_2^T J^2_p + (r_2 \times n_2)^T J^2_w,

        where :math:`n_{1,2}` are normal vectors (note that :math:`n_1 =
        -n_2`), :math:`r_{1, 2}` are vectors from joint origin and nearest
        point, :math:`J^{1, 2}_{p, w}` are position/orientation Jacobians of
        respective joint.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix
                :math:`\frac{\partial h}{\partial q}(q)`.
        """
        # Get robot and collision models and datas
        model = configuration.model
        data = configuration.data
        collision_model = configuration.collision_model
        collision_data = configuration.collision_data

        J = np.zeros((self.dim, model.nv))

        # Calculate `dim` closest collision pairs, and evaluate them
        N_collision = len(collision_model.collisionPairs)
        distances = np.array(
            [
                collision_data.distanceResults[i].min_distance
                for i in range(N_collision)
            ]
        )
        closest_pairs_idxs = np.argpartition(-distances, -self.dim)[
            -self.dim :
        ]

        for i in range(self.dim):
            # Index of the pair
            k = int(closest_pairs_idxs[i])

            cp = collision_model.collisionPairs[k]
            dr = collision_data.distanceResults[k]

            go_1 = collision_model.geometryObjects[cp.first]
            go_2 = collision_model.geometryObjects[cp.second]

            j1_id = go_1.parentJoint
            j2_id = go_2.parentJoint

            w1 = np.array(dr.getNearestPoint1())
            w2 = np.array(dr.getNearestPoint2())
            # Vectors from frame origin to nearest points
            r1 = np.array(w1 - data.oMi[j1_id].translation)
            r2 = np.array(w2 - data.oMi[j2_id].translation)

            # If division by zero is possible, then the points are practically
            # collisind, jacobian is undefined. Set it to zero.
            if np.allclose(w1, w2):
                continue

            # Normal vector betwee nearest points (n_1 in the notation above)
            n = (w1 - w2) / np.linalg.norm(w1 - w2)

            # Calculate first two terms using first frame Jacobian
            J_1 = pin.getJointJacobian(
                model, data, j1_id, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
            )
            Jrow_v = n.T @ J_1[:3, :] + (pin.skew(r1) @ n).T @ J_1[3:, :]

            # Calculate second two terms using second frame Jacobian
            # Note that minus appears, since n_2 = -n_1
            J_2 = pin.getJointJacobian(
                model, data, j2_id, pin.ReferenceFrame.LOCAL_WORLD_ALIGNED
            )
            Jrow_v -= n.T @ J_2[:3, :] + (pin.skew(r2) @ n).T @ J_2[3:, :]

            J[i] = Jrow_v.copy()

        # If collision is undefined, or during the collision, some values
        # might be nans. In this case, set them to zero. Note that for
        # non-colliding smooth convex functions no nans are present.
        J = np.nan_to_num(J)

        return J



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0

"""Test fixture for the joint velocity task."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.exceptions import TargetNotSet, TaskDefinitionError
from pink.tasks import JointVelocityTask


class TestJointVelocityTask(unittest.TestCase):
    """Test consistency of the joint velocity task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description(
            "ur3_description",
            root_joint=pin.JointModelFreeFlyer(),
        )
        configuration = Configuration(robot.model, robot.data, robot.q0)
        self.configuration = configuration
        self.dt = 3e-3  # seconds
        self.nq = configuration.model.nq
        self.nv = configuration.model.nv

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        task = JointVelocityTask(cost=1.0)
        self.assertTrue("cost=" in repr(task))

    def test_target_not_set(self):
        """Exception raised when the target is unset."""
        task = JointVelocityTask(cost=1.0)
        with self.assertRaises(TargetNotSet):
            task.compute_error(self.configuration)

    def test_target_dim_mismatch(self):
        """Exception raised when the target dimension is unexpected."""
        task = JointVelocityTask(cost=1.0)
        task.set_target(np.zeros(self.nv), self.dt)
        with self.assertRaises(TaskDefinitionError):
            task.compute_error(self.configuration)

    def test_matrix_shapes(self):
        """Check the shapes of the error vector and Jacobian matrix."""
        task = JointVelocityTask(cost=1.0)
        task.set_target(np.zeros(self.nv - 6), self.dt)
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        self.assertEqual(e.shape[0], J.shape[0])

    def test_dt(self):
        """Check the shapes of the error vector and Jacobian matrix."""
        task = JointVelocityTask(cost=1.0)

        dt_1 = 3e-3  # seconds
        task.set_target(np.ones(self.nv - 6), dt_1)
        e_1 = task.compute_error(self.configuration)
        self.assertAlmostEqual(e_1[0], dt_1)  # unit target

        dt_2 = 7e-3  # seconds
        task.set_target(np.ones(self.nv - 6), dt_2)
        e_2 = task.compute_error(self.configuration)
        self.assertAlmostEqual(e_2[0], dt_2)  # unit target



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Ivan Domrachev, Simeon Nedelchev

"""Universal Robots UR5 arm tracking a moving target."""

import argparse

import numpy as np
import qpsolvers

import meshcat_shapes
import pink
from pink import solve_ik
from pink.barriers import PositionBarrier
from pink.tasks import FrameTask, PostureTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc

try:
    from robot_descriptions.loaders.pinocchio import load_robot_description
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples need robot_descriptions, "
        "try `[conda|pip] install robot_descriptions`"
    ) from exc  # noqa: E501


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--verbose",
        "-v",
        help="print out task errors and CBF values during execution",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()
    robot = load_robot_description("ur5_description", root_joint=None)
    viz = start_meshcat_visualizer(robot)

    end_effector_task = FrameTask(
        "ee_link",
        position_cost=10.0,  # [cost] / [m]
        orientation_cost=1.0,  # [cost] / [rad]
    )

    posture_task = PostureTask(
        cost=1e-3,  # [cost] / [rad]
    )

    pos_barrier = PositionBarrier(
        "ee_link",
        indices=[1],
        p_min=np.array([-0.4]),
        p_max=np.array([0.6]),
        gain=np.array([100.0]),
        safe_displacement_gain=1.0,
    )
    barriers = [pos_barrier]

    tasks = [end_effector_task, posture_task]

    q_ref = np.array(
        [
            1.27153374,
            -0.87988708,
            1.89104795,
            1.73996951,
            -0.24610945,
            -0.74979019,
        ]
    )
    configuration = pink.Configuration(robot.model, robot.data, q_ref)
    for task in tasks:
        task.set_target_from_configuration(configuration)
    viz.display(configuration.q)

    viewer = viz.viewer
    meshcat_shapes.frame(viewer["end_effector_target"], opacity=0.5)
    meshcat_shapes.frame(viewer["end_effector"], opacity=1.0)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "osqp" in qpsolvers.available_solvers:
        solver = "osqp"

    rate = RateLimiter(frequency=200.0)
    dt = rate.period
    t = 0.0  # [s]
    while True:
        # Update task targets
        end_effector_target = end_effector_task.transform_target_to_world
        end_effector_target.translation[1] = 0.0 + 0.65 * np.sin(t / 4)
        end_effector_target.translation[2] = 0.5

        # Update visualization frames
        viewer["end_effector_target"].set_transform(end_effector_target.np)
        viewer["end_effector"].set_transform(
            configuration.get_transform_frame_to_world(
                end_effector_task.frame
            ).np
        )

        velocity = solve_ik(
            configuration,
            tasks,
            dt,
            solver=solver,
            barriers=barriers,
        )
        configuration.integrate_inplace(velocity, dt)

        G, h = pos_barrier.compute_qp_inequalities(configuration, dt=dt)
        distance_to_manipulator = configuration.get_transform_frame_to_world(
            "ee_link"
        ).translation[1]
        if args.verbose:
            print(
                f"Task error: {end_effector_task.compute_error(configuration)}"
            )
            print(
                "Position CBF value: "
                f"{pos_barrier.compute_barrier(configuration)[0]:0.3f} >= 0"
            )
            print(f"Distance to manipulator: {distance_to_manipulator} <= 0.6")
            print("-" * 60)

        # Visualize result at fixed FPS
        viz.display(configuration.q)
        rate.sleep()
        t += dt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria

"""Omniwheel task implementation."""

import numpy as np

from ..configuration import Configuration
from .rolling_task import RollingTask


class OmniwheelTask(RollingTask):
    r"""Roll without slipping on a plane, allowing lateral motion.

    The name of this task comes from omniwheels, also known as poly wheels.
    See the :class:`pink.tasks.rolling_task.RollingTask`.

    Attributes:
        hub_frame: Name of a frame attached to the hub of the wheel in the
            robot model.
        floor_frame: Name of the inertial frame whose xy-plane defines the
            contact surface the wheel is rolling onto.
        wheel_radius: Radius of the wheel, i.e. distance in meters from the hub
            to the nearest point on the rim.

    Note:
        See also the similar `wheel task in the PlaCo C++ library
        <https://github.com/Rhoban/placo/blob/9e65abadff45071ab11274eb74770d71e10b7ca8/docs/kinematics/wheel_task.rst>`__.
    """

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the rolling task error.

        The error is a 2D vector :math:`{}_R e(q)` of the rim frame :math:`R`:

        .. math::

            {}_R e(q) := \begin{bmatrix}
                0 \\
                z_{\mathit{hub}} - \rho
                \end{bmatrix}

        The error is zero along the x-axis of the rim frame to roll without
        slipping (*i.e.*, the velocity of the contact point is zero in the
        floor plane). The error along the z-axis of the rim frame is chosen so
        that the task keeps the wheel hub at distance :math:`\rho` (wheel
        radius) from the floor plane.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task error :math:`{}_R e(q) \in \mathbb{R}^3`, a translation
            expressed in the rim frame :math:`R`.
        """
        rolling_error = super().compute_error(configuration)
        return np.array([rolling_error[0], rolling_error[2]])

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the omniwheel task Jacobian.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian matrix :math:`{}_R J_{WR}`, expressed locally in the rim
            frame :math:`R`, of shape :math:`(2, n_q)`.
        """
        rolling_jacobian = super().compute_jacobian(configuration)
        return np.vstack([rolling_jacobian[0], rolling_jacobian[2]])

    def __repr__(self):
        """Human-readable representation of the task."""
        rolling_repr = super().__repr__()
        return rolling_repr.replace("RollingTask", "OmniwheelTask")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron
# Copyright 2024 Inria

"""Example of differential IK stuck in a local minimum.

We set up a simple pendulum with joint limits (0, 2 * pi) radians. That is,
joint angles have to be positive, but the joint can do a full turn. We then
define the goal configuration with an angle of 5.5 rad, and initialize the
robot at joint angle 0.5 rad. We task the pendulum with moving its tip to the
tip-position of the goal configuration.

Because differential IK is a local (think "greedy") algorithm, it will move the
tip on the shortest path to target by turning the pendulum clockwise. This will
locally improve the task error, but eventually the pendulum will just get stuck
at its configuration limit (joint angle = 0 rad). At this stage, differential
IK will just keep pushing against the constraint, which is locally the least
worst thing to do.

The global solution would be to accept a temporary increase in the "current tip
to goal tip" distance, and make the pendulum turn anti-clockwise.
"""

import os

import meshcat_shapes
import numpy as np
import pinocchio as pin
import qpsolvers

import pink
from pink import solve_ik
from pink.tasks import FrameTask
from pink.visualization import start_meshcat_visualizer

try:
    from loop_rate_limiters import RateLimiter
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Examples use loop rate limiters, "
        "try `[conda|pip] install loop-rate-limiters`"
    ) from exc


if __name__ == "__main__":
    # Load robot description from robots/simple_pendulum.urdf
    robot = pin.RobotWrapper.BuildFromURDF(
        filename=os.path.join(
            os.path.dirname(__file__),
            "robots",
            "simple_pendulum.urdf",
        ),
        package_dirs=["."],
        root_joint=None,
    )

    # Initialize visualization
    visualizer = start_meshcat_visualizer(robot)
    viewer = visualizer.viewer
    meshcat_shapes.frame(viewer["goal"], opacity=0.5)
    meshcat_shapes.frame(viewer["tip"], opacity=1.0)

    model = robot.model
    data = robot.data
    task = FrameTask(
        "tip",
        position_cost=1.0,  # [cost] / [m]
        orientation_cost=0.61,  # [cost] / [rad]
    )
    tasks = [task]

    # Feasible goal configuration on one side of configuration limits
    q_goal = np.array([5.5])
    goal_configuration = pink.configuration.Configuration(model, data, q_goal)
    goal_pose = goal_configuration.get_transform_frame_to_world("tip")
    viewer["goal"].set_transform(goal_pose.np)
    task.set_target_from_configuration(goal_configuration)  # <= task target

    # Initial configuration on the other side of configuration limits
    q_init = np.array([0.5])
    configuration = pink.configuration.Configuration(model, data, q_init)
    init_pose = configuration.get_transform_frame_to_world("tip")
    viewer["tip"].set_transform(init_pose.np)
    visualizer.display(configuration.q)

    # Select QP solver
    solver = qpsolvers.available_solvers[0]
    if "proxqp" in qpsolvers.available_solvers:
        solver = "proxqp"

    # Run closed-loop inverse kinematics
    rate = RateLimiter(frequency=100.0, warn=False)
    dt = rate.period
    while True:
        velocity = solve_ik(configuration, tasks, dt, solver=solver)
        configuration.integrate_inplace(velocity, dt)
        visualizer.display(configuration.q)
        rate.sleep()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 Inria

"""Joint velocity task."""

from typing import Optional

import numpy as np

from ..configuration import Configuration
from ..exceptions import TargetNotSet, TaskDefinitionError
from ..utils import get_root_joint_dim
from .task import Task


class JointVelocityTask(Task):
    r"""Track joint velocities.

    This task minimizes :math:`\| v^{\mathit{ref}} - v \|_2` with :math:`v` the
    joint velocity resulting from differential IK and :math:`v^{\mathit{ref}}`
    is a reference joint-velocity vector.

    Note:
        Like the :class:`PostureTask`, this task does not apply to
        floating-base coordinates if the robot model has a floating base
        (a.k.a. root joint in Pinocchio).
    """

    def __init__(self, cost: float) -> None:
        r"""Initialize task.

        Args:
            cost: Joint angular velocity cost, in
                :math:`[\mathrm{cost}] [\mathrm{s}] / [\mathrm{rad}]`.
        """
        super().__init__(
            cost=cost,
            gain=1.0,  # unit gain as the task error is directly a velocity
            lm_damping=0.0,  # no Levenberg-Marquardt damping
        )
        self.__target_Delta_q: Optional[np.ndarray] = None

    def set_target(self, target_v: np.ndarray, dt: float) -> None:
        """Set target joint velocity.

        Args:
            target_v: Target joint-velocity vector in the tangent space.
            dt: Integration duration in [s].
        """
        if target_v.ndim != 1:
            raise TaskDefinitionError(
                "joint velocity target should be a vector, "
                f"but the provided target has shape {target_v.shape}"
            )
        self.__target_Delta_q = target_v * dt

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the joint-velocity task error.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Joint-velocity task error.
        """
        _, root_nv = get_root_joint_dim(configuration.model)
        task_nv = configuration.model.nv - root_nv
        if self.__target_Delta_q is None:
            raise TargetNotSet(repr(self))
        if self.__target_Delta_q.shape[0] != task_nv:
            raise TaskDefinitionError(
                f"Target has dimension nv={self.__target_Delta_q.shape[0]} "
                f"but the task expects nv={task_nv} "
                f"({configuration.model.nv=}, {root_nv=})"
            )
        return self.__target_Delta_q

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the joint-velocity task Jacobian.

        The task Jacobian is the identity :math:`I_{n_v} \in \mathbb{R}^{n_v
        \times n_v}`, with :math:`n_v` the dimension of the robot's tangent
        space, so that the task dynamics are:

        .. math::

            J(q) \Delta q = \Delta q

        Combining this with :math:`e(q) = \Delta q_{ref}` and :math:`\alpha =
        1`, we get an overall velocity-tracking task dynamics:

        .. math::

            J(q) \Delta q = -\alpha e(q)
            \quad \Leftrightarrow \quad
            \Delta q = \Delta q_{ref}

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Joint-velocity task Jacobian :math:`J(q)`.
        """
        _, root_nv = get_root_joint_dim(configuration.model)
        return configuration.tangent.eye[root_nv:, :]

    def __repr__(self):
        """Human-readable representation of the task."""
        return f"JointVelocityTask(cost={self.cost})"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2022 Stéphane Caron

"""Test fixture for the joint coupling task."""

import unittest

import numpy as np
import pinocchio as pin
from robot_descriptions.loaders.pinocchio import load_robot_description

from pink import Configuration
from pink.tasks import JointCouplingTask


class TestJointCouplingTask(unittest.TestCase):
    """Test consistency of the joint coupling task.

    Note:
        This fixture only tests the task itself. Integration tests with the IK
        are carried out in :class:`TestSolveIK`.
    """

    def setUp(self):
        """Prepare test fixture."""
        robot = load_robot_description(
            "draco3_description", root_joint=pin.JointModelFreeFlyer()
        )
        self.configuration = Configuration(robot.model, robot.data, robot.q0)

    def test_task_repr(self):
        """String representation reports the task gain, costs and target."""
        task = JointCouplingTask(
            joint_names=["r_knee_fe_jp", "r_knee_fe_jd"],
            ratios=[1.0, -1.0],
            cost=1.0,
            configuration=self.configuration,
        )
        self.assertTrue("cost=" in repr(task))
        self.assertTrue("gain=" in repr(task))

    def test_unit_cost_qp_objective(self):
        """A unit cost vector means the QP objective is (J^T J, -e^T J)."""
        task = JointCouplingTask(
            joint_names=["r_knee_fe_jp", "r_knee_fe_jd"],
            ratios=[1.0, -1.0],
            cost=1.0,
            configuration=self.configuration,
        )
        e = task.compute_error(self.configuration)
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        self.assertTrue(np.allclose(J.T @ J, H))
        self.assertTrue(np.allclose(-e.T @ J, c))

    def test_zero_cost_same_as_disabling_task(self):
        """The task has no effect when its cost is zero."""
        task = JointCouplingTask(
            joint_names=["r_knee_fe_jp", "r_knee_fe_jd"],
            ratios=[1.0, -1.0],
            cost=0.0,
            configuration=self.configuration,
        )
        J = task.compute_jacobian(self.configuration)
        H, c = task.compute_qp_objective(self.configuration)
        qd = np.random.random(J.shape[1:])
        cost = qd.T @ H @ qd + c @ qd
        self.assertAlmostEqual(cost, 0.0)
